/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/

#include "dc/mwv207_panel.h"
#include "mwv207_vi.h"
#include "mwv207_edp.h"
#include <drm/display/drm_dp_helper.h>
#include "mwv207_dp.h"
#include "mwv207_drv.h"
#include "mwv207_vbios.h"
#include <linux/kthread.h>
#include <linux/freezer.h>
#include <linux/spinlock.h>
#include <linux/wait.h>

struct mwv207_edp_config {
	u8 pre_emphasis[4];
	u8 voltage_swing[4];
};

struct mwv207_edp_poll {
	struct task_struct *poll_task;
	wait_queue_head_t poll_wait;
	spinlock_t hpd_lock;
	u8 wait_cond;
};

struct mwv207_edp {
	struct mwv207_output base;

	void __iomem    *mmio;
	int             irq;
	struct completion aux_comp;
	struct mutex aux_mutex;
	struct mwv207_edp_config phy_conf;
	struct mwv207_edp_poll poll;
	struct drm_dp_aux aux;
	struct mwv207_dp_link link;
};

#define connector_to_edp(conn) container_of(conn, struct mwv207_edp, base.connector)
#define encoder_to_edp(encoder) container_of(encoder, struct mwv207_edp, base.encoder)
#define aux_to_edp(aux) container_of(aux, struct mwv207_edp, aux)
#define link_to_edp(link) container_of(link, struct mwv207_edp, link)

#define MWV207_EDP_BASE(idx)   (0x01380000 + 0x40000 * (idx))

static inline u32 mwv207_edp_readb(struct mwv207_edp *edp, u32 reg)
{
	return readl(edp->mmio + reg);
}

static inline void mwv207_edp_writeb(struct mwv207_edp *edp, u32 reg, u32 value)
{
	writel_relaxed(value, edp->mmio + reg);
}

static s32 mwv207_edp_hw_pre_emphasis_get(struct mwv207_edp *edp,
									struct mwv207_dp_link *link, u32 lanenum)
{
	u8 lev = link->train.request.pre_emphasis[lanenum];

	return edp->phy_conf.pre_emphasis[lev];
}
static s32 mwv207_edp_hw_vol_sw_get(struct mwv207_edp *edp,
							struct mwv207_dp_link *link, u32 lanenum)
{
	u8 lev = link->train.request.voltage_swing[lanenum];

	return edp->phy_conf.voltage_swing[lev];
}

static u32 mwv207_edptx_tu_config(u8 lane_count, u32 rate, u32 pixclock)
{
	u32 pixel_clk = pixclock;
	u8 kdefault_tu_size = 64;
	u8 bits_per_color = 8;
	u8 color_factor = 3;
	u64 link_sr, vid_sr;
	u64 data_per_tu;
	u32 a, b;
	u32 tu;

	rate = rate / 100;
	link_sr =
	    (u64) pixel_clk * color_factor * bits_per_color * kdefault_tu_size *
	    10;
	vid_sr = (u64) lane_count * rate * 8;
	data_per_tu = link_sr / vid_sr;

	a = (u32) data_per_tu / 1000;
	b = (u32) (data_per_tu % 1000) * 16 / 1000;
	tu = (b << 24) | (a << 16) | 0x40;

	DRM_DEBUG_DRIVER("EDP pixel_clk = %d khz, rate = %d, lane_count = %d",
			 pixel_clk, rate, lane_count);
	DRM_DEBUG_DRIVER("link_sr = %llu, vid_sr = %llu, data_per_tu = %llu",
			 link_sr, vid_sr, data_per_tu);
	DRM_DEBUG_DRIVER("a = %d, b = %d, tu size: 0x%x", a, b, tu);
	return tu;
}

static int mwv207_edp_phy_parse_cfg(struct mwv207_edp *edp)
{
	u8 pre_emphasis[4] = {0x1, 0x2, 0x3, 0x4};
	u8 voltage_swing[4] = {0x1f, 0x18, 0x1c, 0x1f};

	memcpy(edp->phy_conf.pre_emphasis, pre_emphasis, sizeof(pre_emphasis));
	memcpy(edp->phy_conf.voltage_swing, voltage_swing, sizeof(voltage_swing));

	return 0;
}

static void mwv207_edp_reply_timeout_handle(struct mwv207_edp *edp)
{
	complete(&edp->aux_comp);
	DRM_DEBUG_DRIVER("edp irq timeout!");
}

static void mwv207_edp_reply_handle(struct mwv207_edp *edp)
{
	complete(&edp->aux_comp);
	DRM_DEBUG_DRIVER("edp irq recv!");
}

static void mwv207_edp_hpd_handle(struct mwv207_edp *edp)
{
	spin_lock(&edp->poll.hpd_lock);
	edp->poll.wait_cond |= 0x4;
	spin_unlock(&edp->poll.hpd_lock);

	wake_up_interruptible(&edp->poll.poll_wait);
}

static irqreturn_t mwv207_edp_irq(int irq, void *dev_id)
{
	struct mwv207_edp *edp = dev_id;
	u8 mask;

	mask = mwv207_edp_readb(edp, (0x1000+0x0130U));

	DRM_DEBUG_DRIVER("edp recv irq %d\n", mask);

	while (mask) {
		if (mask & (0x1 << 0)) {

			mwv207_edp_hpd_handle(edp);
		}

		if (mask & (0x1 << 2)) {

			mwv207_edp_reply_handle(edp);
		}

		if (mask & (0x1 << 3)) {

			mwv207_edp_reply_timeout_handle(edp);
		}

		mwv207_edp_readb(edp, (0x1000+0x0140U));
		mask = mwv207_edp_readb(edp, (0x1000+0x0130U));
	}

	return IRQ_HANDLED;
}

static u8 mwv207_edp_get_wait_cond(struct mwv207_edp *edp)
{
	unsigned long flags;
	u8 wait_cond;

	spin_lock_irqsave(&edp->poll.hpd_lock, flags);
	wait_cond = edp->poll.wait_cond;
	edp->poll.wait_cond = 0;
	spin_unlock_irqrestore(&edp->poll.hpd_lock, flags);

	return wait_cond;
}

static int mwv207_edp_poll_task(void *arg)
{
	struct mwv207_edp *edp  = (struct mwv207_edp *) arg;
	int ret;
	u8 wait_cond;

	set_freezable();
	while (!kthread_should_stop()) {
		try_to_freeze();
		ret = wait_event_interruptible_timeout(edp->poll.poll_wait,
						       (wait_cond =
						       mwv207_edp_get_wait_cond
						       (edp)),
						       msecs_to_jiffies(1000));
		if (ret == 0)
			continue;

		if (wait_cond & 0x4) {
			drm_helper_hpd_irq_event(&edp->base.jdev->base);
		}
	}
	return 0;
}

static u32 mwv207_edptx_get_gtcconfig(u32 fapb)
{
	u32 n, b, d;
	s32 a;

	n = 1000 / fapb;
	a = (s32) n;
	b = n - a;
	d = b * (1 << 16);
	d = a << 16 | d;
	return d;
}

static u32 mwv207_edp_get_link_speed(u32 edp_chan, u32 rate)
{
	switch (rate) {
	case 162000:
		return 0x06;
	case 270000:
		return 0x0a;
	case 540000:
		return 0x14;
	}

	DRM_ERROR("edp(%d) rate(%d) didn't find speed", edp_chan, rate);
	return 0;
}

static void mwv207_edp_phy_config(struct mwv207_dp_link *link, u8 lane_count,
			   u32 rate)
{
	unsigned int edp_pll_value[3][8] = {
		{ 0x140067F2, 0x6C8024A6, 0x21653330, 0x474D900E, 0x9C1143D1,
		 0x1, 0x0, 0x1143 },
		{ 0x140067F2, 0x6C8024A6, 0x21653330, 0x474D900E, 0x9C1170D1,
		 0x1, 0x0, 0x1170 },
		{ 0x140067F2, 0x6C8024A6, 0x21653330, 0x47408110, 0x9C0570D1,
		 0x1, 0x0, 0x1070 },
	};
	struct mwv207_edp *edp = link_to_edp(link);
	u32 value;
	u8 rate_inx = 1;
	int status;

	mwv207_edp_writeb(edp, 0x2048, 0);

	mwv207_edp_writeb(edp, 0x2040, 0);

	mwv207_edp_writeb(edp, 0x204C, 0);
	usleep_range(1, 2);

	switch (lane_count) {
	case 1:
		value = 0xf0 | 0x1;
		break;
	case 2:
		value = 0xf0 | 0x3;
		break;
	case 3:
		value = 0xf0 | 0x7;
		break;
	case 4:
		value = 0xf0 | 0xf;
		break;
	default:
		DRM_ERROR("lane_count(%d) num err\n", lane_count);
		return;
	}

	mwv207_edp_writeb(edp, 0x2044, value);
	usleep_range(1, 2);

	mwv207_edp_writeb(edp, 0x2048, 3);
	usleep_range(1, 2);

	value = 0xD9;
	mwv207_edp_writeb(edp, 0x2380, value);
	mwv207_edp_writeb(edp, 0x2384, value);
	mwv207_edp_writeb(edp, 0x2388, value);
	mwv207_edp_writeb(edp, 0x238C, value);

	mwv207_edp_writeb(edp, 0x23B8, 0);
	mwv207_edp_writeb(edp, 0x20B4, 0);
	if (0) {
		mwv207_edp_writeb(edp, 0x23B0,
			   rate == 162000 ? 0x5528 : 0x6118);
	} else {
		mwv207_edp_writeb(edp, 0x23B0, 0);
	}
	mwv207_edp_writeb(edp, 0x20AC, 0);
	mwv207_edp_writeb(edp, 0x23A8, 0);

	if (rate > 270000)
		rate_inx = 2;
	else if (rate < 270000)
		rate_inx = 0;
	else
		rate_inx = 1;

	mwv207_edp_writeb(edp, 0x23A4, edp_pll_value[rate_inx][0]);
	mwv207_edp_writeb(edp, 0x23A0, edp_pll_value[rate_inx][1]);
	mwv207_edp_writeb(edp, 0x239C, edp_pll_value[rate_inx][2]);
	mwv207_edp_writeb(edp, 0x2398, edp_pll_value[rate_inx][3]);
	mwv207_edp_writeb(edp, 0x2394, edp_pll_value[rate_inx][4]);
	mwv207_edp_writeb(edp, 0x20A8, edp_pll_value[rate_inx][5]);
	mwv207_edp_writeb(edp, 0x20A0, edp_pll_value[rate_inx][6]);
	mwv207_edp_writeb(edp, 0x20A4, edp_pll_value[rate_inx][7]);
	usleep_range(10, 11);

	while (time_before(jiffies, msecs_to_jiffies(1))) {
		status = mwv207_edp_readb(edp, 0x2054) & 0x1;
		if (status)
			break;
		usleep_range(10, 20);
	}

	if (!status)
		DRM_DEBUG_DRIVER("edp pll lock timeout!");

	mwv207_edp_writeb(edp, 0x2040, 1);

	switch (lane_count) {
	case 1:
		value = 0x1;
		break;
	case 2:
		value = 0x3;
		break;
	case 3:
		value = 0x7;
		break;
	case 4:
		value = 0xf;
		break;
	default:
		DRM_ERROR("invalid lanes\n");
		return;
	}
	mwv207_edp_writeb(edp, 0x204C, value);
	usleep_range(1, 2);

	mwv207_edp_writeb(edp, 0x23A8, 0x1 << 19);

	mwv207_edp_writeb(edp, 0x2050, 1);
}

static void mwv207_edp_train_config_link_and_phy(struct mwv207_dp_link *link)
{
	struct mwv207_edp *edp = link_to_edp(link);
	u32 ichan = 0;
	u8 speed;

	speed = mwv207_edp_get_link_speed(ichan, link->rate);

	mwv207_edp_writeb(edp, (0x1000+0x0000U), speed);
	mwv207_edp_writeb(edp, (0x1000+0x0004U), link->lanes);

	mwv207_edp_writeb(edp, (0x1000+0x0008U),
		   link->caps.enhanced_framing);
	usleep_range(400, 1000);
	mwv207_edp_writeb(edp, (0x1000+0x001CU),
		   link->caps.alternate_scrambler_reset);

	drm_dp_dpcd_writeb(&edp->aux, DP_LINK_BW_SET, speed);
	drm_dp_dpcd_writeb(&edp->aux, DP_LANE_COUNT_SET,
			   (s8) (link->lanes | 0x80));
	drm_dp_dpcd_writeb(&edp->aux, DP_SET_POWER, 1);

	mwv207_edp_phy_config(link, link->lanes, link->rate);
}

static int mwv207_dp_configure(struct mwv207_dp_link *link)
{
	struct mwv207_edp *edp = link_to_edp(link);

	mwv207_edp_writeb(edp, (0x1000+0x0018U), 0);

	mwv207_edp_train_config_link_and_phy(link);
	DRM_DEBUG_DRIVER("configure lane = %d, rate = %d", link->lanes,
			 link->rate);
	return 0;
}

static int mwv207_dp_apply_training(struct mwv207_dp_link *link)
{
	struct mwv207_edp *edp = link_to_edp(link);
	unsigned int lane;
	u32 voltage_swing = 0;
	u32 pre_emphasis = 0;
	u32 value;

	for (lane = 0; lane < link->lanes; lane++) {
		pre_emphasis = mwv207_edp_hw_pre_emphasis_get(edp, link, lane);
		voltage_swing = mwv207_edp_hw_vol_sw_get(edp, link, lane);

		value =
		    (u32) ((pre_emphasis << 6) | (voltage_swing << 1) | 0x1);
		mwv207_edp_writeb(edp, 0x2380 + 0x4 * lane,
			   value);
		mwv207_edp_writeb(edp, 0x2080 + 0x4 * lane,
			   voltage_swing);
		mwv207_edp_writeb(edp, 0x2090 + 0x4 * lane,
			   pre_emphasis);
		DRM_DEBUG_DRIVER
		    ("lane_%d pre_emphasis = 0x%x, voltage_swing = 0x%x", lane,
		     pre_emphasis, voltage_swing);
	}

	mwv207_edp_writeb(edp, (0x1000+0x000CU),
		   link->train.pattern);
	usleep_range(20, 100);
	DRM_DEBUG_DRIVER("training lane = %d, rate = %d, pattern = %d",
			 link->lanes, link->rate, link->train.pattern);

	return 0;
}

static const struct mwv207_dp_link_ops mwv207_dp_link_funcs = {
	.apply_training = mwv207_dp_apply_training,
	.configure = mwv207_dp_configure,
};

static void mwv207_edp_set_aux_clk(struct mwv207_edp *edp)
{
	int ret;
	int freq_khz;
	int aux_clk;
	u8 value;

	struct mwv207_device *jdev = edp->base.jdev;
	ret = mwv207_vbios_get_pll(jdev, MWV207_PLL_GPU, &freq_khz);
	if (ret < 0)
		DRM_ERROR("mwv207_vbios_get_pll failed");

	aux_clk = freq_khz / 5 / 1000;

	mwv207_edp_writeb(edp, (0x1000+0x010CU), aux_clk);

	value = mwv207_edptx_get_gtcconfig(aux_clk);
	mwv207_edp_writeb(edp, (0x1000+0x0354U), value);
}

static void mwv207_edp_hw_init(struct mwv207_edp *edp)
{

	jdev_modify(edp->base.jdev, 0x9B0920, 0x3 << (20 + edp->base.idx * 4), 1 << (20 + edp->base.idx * 4));

	mwv207_edp_set_aux_clk(edp);

	mwv207_edp_writeb(edp, (0x1000+0x0110U), 400);

	mwv207_edp_writeb(edp, (0x1000+0x0144U), 0x0);
}

static int mwv207_process_aux_ch(struct mwv207_edp *edp, u32 address, u8 *reply,
			    u8 *buffer, u32 size, u32 command)
{
	s32 i;
	unsigned long time_left;
	s32 ret = size;
	s32 err;
	u8 code_reply;
	bool is_write = (command == DP_AUX_NATIVE_WRITE) || (command == DP_AUX_I2C_WRITE);

	err = mutex_lock_interruptible(&edp->aux_mutex);
	if (err)
		return err;

	mwv207_edp_writeb(edp, (0x1000+0x0108U), address);

	if (is_write) {
		for (i = 0; i < size; i++) {
			mwv207_edp_writeb(edp, (0x1000+0x0104U),
				   buffer[i]);
		}
	}

	reinit_completion(&edp->aux_comp);
	if (size == 0) {
		mwv207_edp_writeb(edp, (0x1000+0x0100U),
			   (command << 8) | (0x1 << 12));
	} else {
		mwv207_edp_writeb(edp, (0x1000+0x0100U),
			   (command << 8) | (size - 1));
	}

	DRM_DEBUG_DRIVER("wait_for_completion");
	time_left = wait_for_completion_timeout(&edp->aux_comp,
						msecs_to_jiffies(300));
	if (!time_left) {
		pr_err("%s: aux timeout,\n", __func__);
		ret = -ETIMEDOUT;
		goto unlock_exit;
	}
	DRM_DEBUG_DRIVER("completion");

	if (!is_write) {
		for (i = 0; i < size; i++) {
			buffer[i] = mwv207_edp_readb(edp, (0x1000+0x0134U));
		}
	}

	code_reply = mwv207_edp_readb(edp, (0x1000+0x0138U));
	if (reply)
		*reply = code_reply;

unlock_exit:
	mutex_unlock(&edp->aux_mutex);
	return ret;
}

static void mwv207_encoder_set_edp_timing(struct mwv207_edp *edp,
										struct drm_display_mode *m)
{
	struct mwv207_dp_link *link = &edp->link;
	u32 value;
	u32 symbol;
	u32 udc;

	if (link->lanes == 0 || link->rate == 0) {
		DRM_ERROR
		    ("invalid lane_count(%d) or rate(%d), please check hw!",
		     link->lanes, link->rate);
		return;
	}

	mwv207_edp_writeb(edp, (0x1000+0x0014U), 0);
	usleep_range(400, 401);
	mwv207_edp_writeb(edp, (0x1000+0x0180U), m->crtc_htotal);
	mwv207_edp_writeb(edp, (0x1000+0x0184U), m->crtc_vtotal);

	value = ((~(m->flags & DRM_MODE_FLAG_PVSYNC ? 1 : 0) & 1) << 1) | (~(m->flags & DRM_MODE_FLAG_PHSYNC ? 1 : 0) & 1);
	mwv207_edp_writeb(edp, (0x1000+0x0188U), value);

	mwv207_edp_writeb(edp, (0x1000+0x018CU),
						m->crtc_hsync_end - m->crtc_hsync_start);
	mwv207_edp_writeb(edp, (0x1000+0x0190U),
						m->crtc_vsync_end - m->crtc_vsync_start);
	mwv207_edp_writeb(edp, (0x1000+0x0194U), m->crtc_hdisplay);
	mwv207_edp_writeb(edp, (0x1000+0x0198U), m->crtc_vdisplay);
	value = m->crtc_htotal - m->crtc_hdisplay - (m->crtc_hsync_start - m->crtc_hdisplay);
	mwv207_edp_writeb(edp, (0x1000+0x019CU), value);

	value = m->crtc_vtotal - m->crtc_vdisplay - (m->crtc_vsync_start - m->crtc_vdisplay);
	mwv207_edp_writeb(edp, (0x1000+0x01A0U), value);

	mwv207_edp_writeb(edp, (0x1000+0x01A8U), 0);
	mwv207_edp_writeb(edp, (0x1000+0x01ACU), m->crtc_clock);
	mwv207_edp_writeb(edp, (0x1000+0x01B0U), 0x40);

	symbol = ((m->crtc_hdisplay * 24) + 7) / 8;
	udc = (symbol + link->lanes - 1) / link->lanes;
	mwv207_edp_writeb(edp, (0x1000+0x01BCU), udc);
	mwv207_edp_writeb(edp, (0x1000+0x01C0U), 0);
	mwv207_edp_writeb(edp, (0x1000+0x01ACU), m->crtc_clock / 10);
	mwv207_edp_writeb(edp, (0x1000+0x01C4U), 0xf);
	mwv207_edp_writeb(edp, (0x1000+0x01B4U), link->rate / 10);

	mwv207_edp_writeb(edp, (0x1000+0x01A4U), 0x21);
	mwv207_edp_writeb(edp, (0x1000+0x01A8U), 0);

	mwv207_edp_writeb(edp, (0x1000+0x01B8U), 1);

	value = mwv207_edptx_tu_config(link->lanes, link->rate, m->crtc_clock);
	mwv207_edp_writeb(edp, (0x1000+0x01B0U), value);
	mwv207_edp_writeb(edp, (0x1000+0x00C8U), 0x20042004);
	mwv207_edp_writeb(edp, (0x1000+0x00CCU), 0x20042004);
	mwv207_edp_writeb(edp, (0x1000+0x01ACU), m->crtc_clock / 10);
	mwv207_edp_writeb(edp, (0x1000+0x01B4U), link->rate / 10);

	mwv207_edp_writeb(edp, (0x1000+0x002CU), 0);
	mwv207_edp_writeb(edp, (0x1000+0x0094U), 0x1);

	mwv207_edp_writeb(edp, (0x1000+0x0084U), 1);
}

static ssize_t mwv207_edp_aux_transfer(struct drm_dp_aux *aux,
				      struct drm_dp_aux_msg *msg)
{
	ssize_t result = 0;
	struct mwv207_edp *edp = aux_to_edp(aux);
	u32 request = msg->request & ~(DP_AUX_I2C_MOT | DP_AUX_I2C_WRITE_STATUS_UPDATE);

	if (WARN_ON(msg->size > 16))
		return -E2BIG;

	switch (request) {
	case DP_AUX_NATIVE_WRITE:
	case DP_AUX_I2C_WRITE:
	case DP_AUX_NATIVE_READ:
	case DP_AUX_I2C_READ:
		result =  mwv207_process_aux_ch(edp, msg->address, &msg->reply,
				 (u8 *) msg->buffer, msg->size, request);
		break;
	default:
		return -EINVAL;
	}

	if (((request == DP_AUX_NATIVE_WRITE)
		|| (request == DP_AUX_I2C_WRITE))
		&& result >= 0)
		result = msg->size;

	DRM_DEBUG_DRIVER("Op: %x, addr: 0x%04x, size: %ld, reply code = %x",
			 request, msg->address, msg->size, msg->reply);
	return result;
}

static void mwv207_edp_destroy(struct drm_connector *conn)
{
	drm_connector_unregister(conn);
	drm_connector_cleanup(conn);
}

static int mwv207_edp_output_late_register(struct drm_connector *connector)
{
	int ret;
	struct mwv207_edp *edp = connector_to_edp(connector);
	struct mwv207_output *output = connector_to_output(connector);

    	mwv207_kms_panel_init(connector->dev,output->jdev,&output->panel,
                          output->idx,output->connector.connector_type);

	edp->aux.name = "mwv207_edp_aux";
	edp->aux.dev = connector->dev->dev;
	edp->aux.drm_dev = connector->dev;
	edp->aux.transfer = mwv207_edp_aux_transfer;
	ret = drm_dp_aux_register(&edp->aux);
	if (ret) {
		DRM_ERROR("failed to register drm aux: %d\n", ret);
		return -EINVAL;
	}
	edp->link.ops = &mwv207_dp_link_funcs;
	edp->link.aux = &edp->aux;
	output->ddc = &edp->aux.ddc;

	return 0;
}

static void mwv207_edp_output_early_unregister(struct drm_connector *connector)
{
	struct mwv207_edp *edp = connector_to_edp(connector);
	drm_dp_aux_unregister(&edp->aux);
}

static enum drm_mode_status mwv207_edp_mode_valid(struct mwv207_edp *edp,
		const struct drm_display_mode *mode)
{

	if (mode->clock > 594000)
		return MODE_CLOCK_HIGH;

	return MODE_OK;
}

static int mwv207_edp_get_modes(struct drm_connector *connector)
{
	return mwv207_output_get_modes(connector);
}

static enum drm_mode_status mwv207_edp_connector_mode_valid(struct drm_connector *connector,
		struct drm_display_mode *mode)
{
	return mwv207_edp_mode_valid(connector_to_edp(connector), mode);
}

static int mwv207_edp_detect_ctx(struct drm_connector *connector,
			  struct drm_modeset_acquire_ctx *ctx,
			  bool force)
{
	struct mwv207_edp *edp = connector_to_edp(connector);
	u32 value;

	value = mwv207_edp_readb(edp, (0x1000+0x0128U));
	if (value)
		return connector_status_connected;
	else
		return connector_status_disconnected;
}

static void mwv207_edp_switch(struct mwv207_output *output, bool on)
{
	mwv207_output_modify(output, MWV207_EDP_CTRL(output->idx), 1 << 3, (on ? 1 : 0) << 3);
}

static void mwv207_edp_select_crtc(struct mwv207_output *output)
{
	jdev_modify(output->jdev, 0x9B0034, 0x1 << (8 + output->idx), 0);
	jdev_modify(output->jdev, 0x9B0038, 0x1 << (8 + output->idx), 0);

	msleep(20);
	jdev_modify(output->jdev, 0x9B0034, 0x1 << (8 + output->idx),
			    0x1 << (8 + output->idx));
	jdev_modify(output->jdev, 0x9B0038, 0x1 << (8 + output->idx),
			    0x1 << (8 + output->idx));
	msleep(20);

	mwv207_output_modify(output, MWV207_EDP_CTRL(output->idx),
			0x7, drm_crtc_index(output->cur_crtc));

	jdev_modify(output->jdev, 0x9B003C, 0b111 << (24 + output->idx * 4),
			drm_crtc_index(output->cur_crtc) << (24 + output->idx * 4));
}

static void mwv207_encoder_edp_config_para(struct mwv207_edp *edp)
{
	mwv207_edp_hw_init(edp);

	mwv207_edp_writeb(edp, (0x1000+0x0090U), 0x01);

	mwv207_edp_writeb(edp, (0x1000+0x0080U), 0x1);

	mwv207_edp_set_aux_clk(edp);

	mwv207_edp_writeb(edp, (0x1000+0x0024U), 0x12);

	mwv207_edp_writeb(edp, (0x1000+0x0014U), 1);
}

static void mwv207_edp_set_timing(struct mwv207_edp *edp)
{
	struct drm_display_mode *mode;
	struct drm_display_info *info;
	int err;
	mode = &edp->base.cur_crtc->state->adjusted_mode;
	info = &edp->base.connector.display_info;

	mwv207_encoder_edp_config_para(edp);

	err = mwv207_dp_link_probe(edp->link.aux, &edp->link);
	if (err < 0)
		DRM_DEBUG_DRIVER("failed to probe DP link: %d\n", err);

	err = mwv207_dp_link_choose(&edp->link, mode, info);
	if (err < 0)
		DRM_DEBUG_DRIVER("failed to choose link: %d\n", err);

	err = mwv207_dp_link_train(&edp->link);
	if (err < 0)
		DRM_DEBUG_DRIVER("link training failed: %d\n", err);

	err = mwv207_dp_link_power_up(edp->link.aux, &edp->link);
	if (err < 0)
		DRM_DEBUG_DRIVER("failed to power up DP link: %d\n", err);

	mwv207_encoder_set_edp_timing(edp, mode);
}

static void mwv207_edp_encoder_disable(struct drm_encoder *encoder)
{
	struct mwv207_output *output = encoder_to_output(encoder);
	struct mwv207_edp *edp = encoder_to_edp(encoder);
	int err;

	mwv207_edp_writeb(edp, (0x1000+0x0084U), 0);
	mwv207_edp_switch(output, false);
        mwv207_panel_disable(output->panel);
        mwv207_panel_unprepare(output->panel);
	err = mwv207_dp_link_power_down(edp->link.aux, &edp->link);
	if (err < 0)
		DRM_DEBUG_DRIVER("failed to power down DP link: %d\n", err);
}

static void mwv207_edp_encoder_reset(struct drm_encoder *encoder)
{
	struct mwv207_edp *edp = encoder_to_edp(encoder);
	mwv207_edp_encoder_disable(encoder);

	mwv207_edp_writeb(edp, (0x1000+0x0080U), 0x1);
}

static enum drm_mode_status mwv207_edp_encoder_mode_valid(struct drm_encoder
							 *encoder, const struct
							 drm_display_mode * mode)
{
	return mwv207_edp_mode_valid(encoder_to_edp(encoder), mode);
}

static int mwv207_edp_encoder_atomic_check(struct drm_encoder *encoder,
					  struct drm_crtc_state *crtc_state,
					  struct drm_connector_state
					  *conn_state)
{
	return 0;
}

static void mwv207_edp_encoder_enable(struct drm_encoder *encoder)
{
	struct mwv207_output *output = encoder_to_output(encoder);

        mwv207_panel_prepare(output->panel);
	mwv207_edp_select_crtc(output);
	mwv207_edp_set_timing(encoder_to_edp(encoder));
	mwv207_edp_switch(output, true);
        mwv207_panel_enable(output->panel);
}

static const struct drm_connector_funcs mwv207_edp_connector_funcs = {
	.reset                  = drm_atomic_helper_connector_reset,
	.fill_modes             = drm_helper_probe_single_connector_modes,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state   = drm_atomic_helper_connector_destroy_state,
	.destroy                = mwv207_edp_destroy,
	.late_register		= mwv207_edp_output_late_register,
	.early_unregister	= mwv207_edp_output_early_unregister,
};

static const struct drm_connector_helper_funcs mwv207_edp_connector_helper_funcs = {
	.get_modes  = mwv207_edp_get_modes,
	.mode_valid = mwv207_edp_connector_mode_valid,
	.detect_ctx = mwv207_edp_detect_ctx,
};

static const struct drm_encoder_funcs mwv207_edp_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
	.reset   = mwv207_edp_encoder_reset,
};

static const struct drm_encoder_helper_funcs mwv207_edp_encoder_helper_funcs = {
	.mode_valid      = mwv207_edp_encoder_mode_valid,
	.atomic_check    = mwv207_edp_encoder_atomic_check,
	.enable          = mwv207_edp_encoder_enable,
	.disable         = mwv207_edp_encoder_disable,
};

static int mwv207_edp_init_single(struct mwv207_device *jdev, int idx)
{
	struct mwv207_output *output;
	struct mwv207_edp *edp;
	int ret;

	edp = devm_kzalloc(jdev->dev, sizeof(*edp), GFP_KERNEL);
	if (!edp)
		return -ENOMEM;

	output = &edp->base;
	output->jdev = jdev;
	output->idx  = idx;
	output->mmio = jdev->mmio + 0x9A0000;
	edp->mmio = jdev->mmio + MWV207_EDP_BASE(idx);

	init_completion(&edp->aux_comp);
	mutex_init(&edp->aux_mutex);
	spin_lock_init(&edp->poll.hpd_lock);
	init_waitqueue_head(&edp->poll.poll_wait);
	edp->poll.wait_cond = 0;
	edp->poll.poll_task =
	    kthread_run(mwv207_edp_poll_task, edp,
			"mwv207_dp_hpd_handler");
	if (IS_ERR(edp->poll.poll_task)) {
		DRM_ERROR("unable to create kernel thread: %ld\n",
			  PTR_ERR(edp->poll.poll_task));
	}

	ret = mwv207_edp_phy_parse_cfg(edp);
	if (ret)
		return ret;

	output->connector.polled = DRM_CONNECTOR_POLL_HPD;
	ret = drm_connector_init(&jdev->base, &output->connector,
			&mwv207_edp_connector_funcs, DRM_MODE_CONNECTOR_DisplayPort);
	if (ret)
		return ret;
	drm_connector_helper_add(&output->connector, &mwv207_edp_connector_helper_funcs);

	output->encoder.possible_crtcs = (1 << jdev->base.mode_config.num_crtc) - 1;
	ret = drm_encoder_init(&jdev->base, &output->encoder,
			&mwv207_edp_encoder_funcs, DRM_MODE_ENCODER_TMDS,
			"edp-%d", output->idx);
	if (ret)
		return ret;

	drm_encoder_helper_add(&output->encoder, &mwv207_edp_encoder_helper_funcs);

	ret = drm_connector_attach_encoder(&output->connector, &output->encoder);
	if (ret)
		return ret;

	edp->irq = irq_find_mapping(jdev->irq_domain, output->idx + 0x28);
	BUG_ON(edp->irq == 0);

	ret = devm_request_irq(jdev->dev, edp->irq, mwv207_edp_irq, 0,
			       output->encoder.name, edp);
	if (ret < 0) {
		dev_err(jdev->dev, "failed to request IRQ#%u: %d\n",
			edp->irq, ret);
		return ret;
	}

	mwv207_edp_hw_init(edp);
	return drm_connector_register(&output->connector);
}

int mwv207_edp_init(struct mwv207_device *jdev)
{
	int i, ret;

	for (i = 0; i < 0x1; i++) {
		ret = mwv207_edp_init_single(jdev, i);
		if (ret)
			return ret;
	}

	pr_info("%s TBD", __func__);
	return 0;
}

