/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/version.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_fb_helper.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_fourcc.h>
#include <drm/drm_print.h>
#include <drm/drm_framebuffer.h>
#include <linux/fb.h>
#include <drm/drm_gem_framebuffer_helper.h>
#include "mwv207_drv.h"
#include "mwv207_bo.h"
#include "mwv207_drm.h"
#include "mwv207_gem.h"
#include "mwv207_kms.h"
#include "mwv207_drv.h"

static struct fb_ops mwv207_fb_ops = {
	.owner          = THIS_MODULE,
	DRM_FB_HELPER_DEFAULT_OPS,
	.fb_fillrect    = cfb_fillrect,
	.fb_copyarea    = cfb_copyarea,
	.fb_imageblit   = cfb_imageblit,
};

static int mwv207_fb_create(struct drm_fb_helper *fb_helper,
		struct drm_fb_helper_surface_size *sizes)
{
	struct drm_mode_fb_cmd2 mode_cmd;
	struct drm_framebuffer *fb;
	struct mwv207_device *jdev;
	struct mwv207_bo *jbo;
	struct fb_info *info;
	u32 bytes;
	void *logical;
	int ret;

	jdev = fb_helper->dev->dev_private;
	memset(&mode_cmd, 0, sizeof(mode_cmd));
	mode_cmd.width = sizes->surface_width;
	mode_cmd.height = sizes->surface_height;
	if (sizes->surface_bpp == 24)
		sizes->surface_bpp = 32;

	mode_cmd.pitches[0] = mode_cmd.width * ((sizes->surface_bpp + 7)/8);
	mode_cmd.pixel_format = drm_mode_legacy_fb_format(sizes->surface_bpp,
			sizes->surface_depth);

	info = drm_fb_helper_alloc_info(fb_helper);
	if (IS_ERR(info))
		return  PTR_ERR(info);

	bytes  = mode_cmd.pitches[0] * mode_cmd.height;
	bytes = ALIGN(bytes, PAGE_SIZE);
	ret = mwv207_bo_create(jdev, bytes, 0x10000, true,
			       0x2,
			       0x1,
			       NULL, NULL, &jbo);
	if (ret)
		return ret;

	fb = kzalloc(sizeof(*fb), GFP_KERNEL);
	if (fb == NULL) {
		ret = -ENOMEM;
		goto free_bo;
	}

	ret = mwv207_framebuffer_init(jdev, fb, &mode_cmd, mwv207_gem_from_bo(jbo));
	if (ret)
		goto free_fb;

	ret = mwv207_bo_reserve(jbo, true);
	if (ret)
		goto free_fb;
	ret = mwv207_bo_pin_reserved(jbo, 0x2);
	if (ret)
		goto unreserve_bo;

	ret = mwv207_bo_kmap_reserved(jbo, &logical);
	if (ret)
		goto unpin_bo;

	memset(logical, 0x0, bytes);

	fb_helper->fb = fb;
	info->skip_vt_switch = true;
	info->fbops = &mwv207_fb_ops;
	info->screen_size = fb->height * fb->pitches[0];
	info->fix.smem_len = info->screen_size;
	info->screen_base = logical;

	drm_fb_helper_fill_info(info, fb_helper, sizes);

	mwv207_bo_unreserve(jbo);
	return 0;

unpin_bo:
	mwv207_bo_unpin(jbo);
unreserve_bo:
	mwv207_bo_unreserve(jbo);
free_fb:
	kfree(fb);
free_bo:
	mwv207_bo_unref(jbo);
	return ret;
}

static const struct drm_fb_helper_funcs mwv207_fb_helper_funcs = {
	.fb_probe = mwv207_fb_create,
};

static void mwv207_fbdev_client_unregister(struct drm_client_dev *client)
{
	struct drm_fb_helper *fb_helper = drm_fb_helper_from_client(client);

	if (fb_helper->info) {
		drm_fb_helper_unregister_info(fb_helper);
	} else {
		drm_client_release(&fb_helper->client);
		drm_fb_helper_unprepare(fb_helper);
		kfree(fb_helper);
	}
}

static int mwv207_fbdev_client_restore(struct drm_client_dev *client)
{
	drm_fb_helper_lastclose(client->dev);

	return 0;
}

static int mwv207_fbdev_client_hotplug(struct drm_client_dev *client)
{
	struct drm_fb_helper *fb_helper = drm_fb_helper_from_client(client);
	struct drm_device *dev = client->dev;
	int ret;

	if (dev->fb_helper)
		return drm_fb_helper_hotplug_event(dev->fb_helper);

	ret = drm_fb_helper_init(dev, fb_helper);
	if (ret)
		goto err_drm_err;

	if (!drm_drv_uses_atomic_modeset(dev))
		drm_helper_disable_unused_functions(dev);

	ret = drm_fb_helper_initial_config(fb_helper);
	if (ret)
		goto err_drm_fb_helper_fini;

	return 0;

err_drm_fb_helper_fini:
	drm_fb_helper_fini(fb_helper);
err_drm_err:
	drm_err(dev, "Failed to setup fbdev emulation (ret=%d)\n", ret);
	return ret;
}

static const struct drm_client_funcs mwv207_fbdev_client_funcs = {
	.owner		= THIS_MODULE,
	.unregister	= mwv207_fbdev_client_unregister,
	.restore	= mwv207_fbdev_client_restore,
	.hotplug	= mwv207_fbdev_client_hotplug,
};

int mwv207_fbdev_init(struct mwv207_device *jdev)
{
	int preferred_bpp = 32;
  struct drm_device *dev = &jdev->base;
	struct mwv207_fbdev *fbdev;
	int ret;

	fbdev = devm_kzalloc(jdev->dev, sizeof(*fbdev), GFP_KERNEL);
	if (!fbdev)
		return -ENOMEM;

	drm_fb_helper_prepare(dev,&fbdev->helper, preferred_bpp, &mwv207_fb_helper_funcs);

	ret = drm_client_init(dev, &fbdev->helper.client, "fbdev", &mwv207_fbdev_client_funcs);
	if (ret)
		goto err_drm_client_init;

	drm_client_register(&fbdev->helper.client);

  fbdev->helper.dev = dev;
  jdev->fbdev = fbdev;
	return 0;
err_drm_client_init:
  drm_fb_helper_fini(&fbdev->helper);
	return ret;
}

void mwv207_fbdev_fini(struct mwv207_device *jdev)
{
	struct mwv207_fbdev *fbdev = jdev->fbdev;

	if (!fbdev)
		return;

	drm_framebuffer_unregister_private(&fbdev->fb);
	pr_info("Module refcount: %u\n", module_refcount(THIS_MODULE));
	drm_framebuffer_cleanup(&fbdev->fb);
	pr_info("Module refcount: %u\n", module_refcount(THIS_MODULE));
	drm_fb_helper_fini(&fbdev->helper);
	pr_info("Module refcount: %u\n", module_refcount(THIS_MODULE));

	if (mwv207_bo_reserve(fbdev->jbo, false) == 0) {
	pr_info("Module refcount: %u\n", module_refcount(THIS_MODULE));
		mwv207_bo_kunmap_reserved(fbdev->jbo);
		mwv207_bo_unpin(fbdev->jbo);
		mwv207_bo_unreserve(fbdev->jbo);
		mwv207_bo_unref(fbdev->jbo);
	}
	pr_info("Module refcount: %u\n", module_refcount(THIS_MODULE));
	jdev->fbdev = NULL;
}
