/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/

#include <drm/display/drm_hdmi_helper.h>
#include <linux/delay.h>
#include "mwv207_drv.h"
#include "mwv207_vi.h"
#include "dw-hdmi.h"
#include "mwv207_vbios.h"

#include <sound/core.h>
#include <sound/asoundef.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/initval.h>
#include <sound/control.h>
#include <sound/jack.h>
#include <sound/pcm_iec958.h>
#include <sound/pcm_drm_eld.h>

#define DRIVER_NAME        "mwv207"

static int hdmi_card_index = SNDRV_DEFAULT_IDX1;
static char *hdmi_card_id = SNDRV_DEFAULT_STR1;
static const struct snd_pcm_hardware had_pcm_hardware;

struct channel_config {
	u8 conf1;
	u8 ca;
};

static struct channel_config default_hdmi_channel_config[7] = {
	{ 0x03, 0x00 },
	{ 0x0b, 0x02 },
	{ 0x33, 0x08 },
	{ 0x37, 0x09 },
	{ 0x3f, 0x0b },
	{ 0x7f, 0x0f },
	{ 0xff, 0x13 },
};

static const struct snd_pcm_hardware had_pcm_hardware = {
	.info = (SNDRV_PCM_INFO_INTERLEAVED |
		 SNDRV_PCM_INFO_BLOCK_TRANSFER |
		 SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_MMAP_VALID),
	.formats = SNDRV_PCM_FMTBIT_S24_LE,
	.rates = (SNDRV_PCM_RATE_32000 |
		  SNDRV_PCM_RATE_44100 |
		  SNDRV_PCM_RATE_48000 |
		  SNDRV_PCM_RATE_88200 |
		  SNDRV_PCM_RATE_96000 |
		  SNDRV_PCM_RATE_176400 | SNDRV_PCM_RATE_192000),
	.channels_min = 0x2,
	.channels_max = 0x8,
	.rate_min = 0x7D00,
	.rate_max = 0x2EE00,
	.buffer_bytes_max = (1024*1024),
	.period_bytes_min = 0x100,
	.period_bytes_max = 0x2000,
	.periods_min = 0x2,
	.periods_max = 0x10,
	.fifo_size = 0x0,
};

enum video_if_reg {
	VIDEO_IF_CTRL = 0x00,
	VIDEO_IF_TMDS_CTRL = 0x04,
	VIDEO_IF_DMA_HADDR_L = 0x08,
	VIDEO_IF_DMA_HADDR_H = 0x0C
};

struct pcm_stream_info {
	struct snd_pcm_substream *substream;
	int substream_refcount;
};

struct port_cfg {
	const char *eld;
	int clock;
	bool enabled;
	bool hw_connected;
};

struct mwv207_audio_port {
	struct mwv207_audio_card *jcard;
	bool is_dp;
	bool running;
	bool enabled;
	bool has_jack;
	struct pcm_stream_info stream_info;

	u8 eld[0x80];
	int id;
	int pipe;

	struct device *dev;

	int clock;

	spinlock_t lock;
	struct mutex mutex;

	u32 buf_offset;
	u32 buf_period;
	u32 buf_size;
	u32 channels;
	dma_addr_t dma_addr;

	struct snd_pcm *pcm;
	struct snd_jack *jack;

	struct hrtimer fake_dma;
	ktime_t period_time;
	int use_fake_dma;

	struct delayed_work jack_work;
	struct port_cfg cfg;
	bool active;
};

struct mwv207_audio_card {
	struct snd_card *card;
	struct device *dev;

	phys_addr_t phys_base;
	phys_addr_t phys_len;
	void __iomem *mmio_base;

	int port_nr;
	struct mwv207_audio_port port[0x3];
};

struct mwv207_audio {
	struct mwv207_audio_card *jcard;
};

static inline u32 had_reg_remap(struct mwv207_audio_card *jcard, u32 reg)
{
	if (unlikely(jcard->phys_len < 0x1200000)) {
		BUG_ON(reg < 0x1200000
		       || reg >= 0x1200000 + 0x200000);
		reg -= 0x1200000;
		reg += 0xA00000;
	}
	return reg;
}

static inline u32
had_read_hdmi_tx_reg_raw(struct mwv207_audio_port *port, u32 offset)
{
	return ioread32(port->jcard->mmio_base +
			had_reg_remap(port->jcard,
				      HDMI_TX_REG(port->id, offset)));
}

static inline u32 had_read_video_if_reg(struct mwv207_audio_port *port, u32 offset)
{
	if (port->is_dp)
		return 0;
	return ioread32(port->jcard->mmio_base +
			VIDEO_IF_REG(port->id, offset));
}

static inline void
had_write_hdmi_tx_reg_raw(struct mwv207_audio_port *port, u32 offset, u32 value)
{
	iowrite32(value, port->jcard->mmio_base +
		  had_reg_remap(port->jcard, HDMI_TX_REG(port->id, offset)));
}

static inline u32 had_read_hdmi_tx_reg(struct mwv207_audio_port *port, u32 offset)
{
	if (!port->enabled)
		return 0;
	return had_read_hdmi_tx_reg_raw(port, offset);
}

static inline void
had_write_video_if_reg(struct mwv207_audio_port *port, u32 offset, u32 value)
{
	if (port->is_dp)
		return;
	iowrite32(value, port->jcard->mmio_base +
		  VIDEO_IF_REG(port->id, offset));
}

static inline void
had_write_hdmi_tx_reg(struct mwv207_audio_port *port, u32 offset, u32 value)
{
	if (port->enabled)
		had_write_hdmi_tx_reg_raw(port, offset, value);
}

static void had_init_dma_haddr(struct mwv207_audio_port *port)
{
	u32 axi_l, axi_h;
	u64 axi_offset;
	if (port->is_dp)
		return;

	axi_offset = 0x800000000UL;

	axi_l = (u32) axi_offset;
	axi_h = (u32) (axi_offset >> 32);

	had_write_video_if_reg(port, VIDEO_IF_DMA_HADDR_L, axi_l);
	had_write_video_if_reg(port, VIDEO_IF_DMA_HADDR_H, axi_h);
}

static void had_stop_dma(struct mwv207_audio_port *port)
{

	had_write_hdmi_tx_reg_raw(port, 0x3614, ~0U);

	had_write_hdmi_tx_reg_raw(port, 0x3602,
				  HDMI_AHB_DMA_STOP_STOP_MASK);
}

static void had_init_dma_interrupt(struct mwv207_audio_port *port)
{
	had_write_hdmi_tx_reg(port, 0x3602,
			      HDMI_AHB_DMA_STOP_STOP_MASK);

	had_init_dma_haddr(port);

	had_write_hdmi_tx_reg(port, 0x109,
			      HDMI_IH_AHBDMAAUD_STAT0_MASK);

	had_write_hdmi_tx_reg(port, 0x189,
			      HDMI_IH_MUTE_AHBDMAAUD_STAT0_MASK &
			      ~HDMI_IH_MUTE_AHBDMAAUD_STAT0_DONE);
}

static void set_auds_sample(struct mwv207_audio_port *port, u32 samp_freq)
{
	u32 val;

	had_write_hdmi_tx_reg(port, 0x1065, 0x00);
	had_write_hdmi_tx_reg(port, 0x1066, 0x00);
	had_write_hdmi_tx_reg(port, 0x1067, 0x00);
	had_write_hdmi_tx_reg(port, 0x1068, 0x00);
	had_write_hdmi_tx_reg(port, 0x1069, 0x02);
	had_write_hdmi_tx_reg(port, 0x106A, 0x02);
	had_write_hdmi_tx_reg(port, 0x106B, 0x00);
	had_write_hdmi_tx_reg(port, 0x106C, 0x01);
	had_write_hdmi_tx_reg(port, 0x106D, 0x00);

	val = (samp_freq == 0xBB80) ? 0xc2 : 0xc0;
	had_write_hdmi_tx_reg(port, 0x106E, val);

	val = (samp_freq == 0xBB80) ? 0xd5 : 0xf5;
	had_write_hdmi_tx_reg(port, 0x106F, val);
}

static void had_set_cts_n(struct mwv207_audio_port *port, u32 cts, u32 n)
{
	u32 value;

	value = had_read_hdmi_tx_reg(port, 0x3205);
	value &= (u32) (~HDMI_AUD_CTS3_CTS_MANUAL);
	had_write_hdmi_tx_reg(port, 0x3205, value);

	value = had_read_hdmi_tx_reg(port, 0x3205);
	value &= (u32) (~HDMI_AUD_CTS3_N_SHIFT_MASK);
	had_write_hdmi_tx_reg(port, 0x3205, value);
	had_write_hdmi_tx_reg(port, 0x3202, ((n >> 16) & 0xff) | 0x80);

	had_write_hdmi_tx_reg(port, 0x3205, ((cts >> 16)
						    &
						    HDMI_AUD_CTS3_AUDCTS19_16_MASK)
			      | HDMI_AUD_CTS3_CTS_MANUAL);
	had_write_hdmi_tx_reg(port, 0x3204, (cts >> 8) & 0xff);
	had_write_hdmi_tx_reg(port, 0x3203, cts & 0xff);

	had_write_hdmi_tx_reg(port, 0x3202, ((n >> 16) & 0xff) | 0x80);
	had_write_hdmi_tx_reg(port, 0x3201, (n >> 8) & 0xff);
	had_write_hdmi_tx_reg(port, 0x3200, n & 0xff);
}

static int had_calculate_maud_value(u32 aud_samp_freq, u32 link_rate)
{
	u32 maud_val;

	if (link_rate == 0x41EB0) {
		switch (aud_samp_freq) {
		case 0x7D00:
			maud_val = 0x7C4;
			break;
		case 0xAC44:
			maud_val = 0xAB4;
			break;
		case 0xBB80:
			maud_val = 0xBA6;
			break;
		case 0x15888:
			maud_val = 0x1568;
			break;
		case 0x17700:
			maud_val = 0x174D;
			break;
		case 0x2B110:
			maud_val = 0x2AD1;
			break;
		case 0x2EE00:
			maud_val = 0x2E9A;
			break;
		default:
			maud_val = -EINVAL;
			break;
		}
	} else if (link_rate == 0x278D0) {
		switch (aud_samp_freq) {
		case 0x7D00:
			maud_val = 0xCF2;
			break;
		case 0xAC44:
			maud_val = 0x11D7;
			break;
		case 0xBB80:
			maud_val = 0x136B;
			break;
		case 0x15888:
			maud_val = 0x23AE;
			break;
		case 0x17700:
			maud_val = 0x26D6;
			break;
		case 0x2B110:
			maud_val = 0x475C;
			break;
		case 0x2EE00:
			maud_val = 0x4DAC;
			break;
		default:
			maud_val = -EINVAL;
			break;
		}
	} else
		maud_val = -EINVAL;

	return maud_val;
}

static void had_program_cts(struct mwv207_audio_port *port, u32 samp_freq,
			    u32 clock, u32 n_param)
{
	u32 cts_val;
	u64 dividend, divisor;
	u32 config3 = 0;

	if (port->is_dp) {

		cts_val = had_calculate_maud_value(samp_freq, clock);
	} else {
		dividend = (u64) clock * n_param * 1000;
		divisor = 128 * samp_freq;
		cts_val = div64_u64(dividend, divisor);
	}

	config3 = had_read_hdmi_tx_reg(port, 0x7);

	if (!(config3 & HDMI_CONFIG3_AHBAUDDMA))
		cts_val = 0;

	had_set_cts_n(port, cts_val, n_param);
}

static int had_calculate_n_value(u32 samp_freq)
{
	int n_val;

	switch (samp_freq) {
	case 0x7D00:
		n_val = 4096;
		break;
	case 0xAC44:
		n_val = 6272;
		break;
	case 0xBB80:
		n_val = 6144;
		break;
	case 0x15888:
		n_val = 12544;
		break;
	case 0x17700:
		n_val = 12288;
		break;
	case 0x2B110:
		n_val = 25088;
		break;
	case 0x2EE00:
		n_val = 24576;
		break;
	default:
		n_val = -EINVAL;
		break;
	}
	return n_val;
}

static int had_program_n(struct mwv207_audio_port *port,
			 u32 samp_freq, u32 *n_param)
{
	int n_val;

	if (port->is_dp) {
		n_val = 0x8000;
	} else {
		n_val = had_calculate_n_value(samp_freq);
	}

	if (n_val < 0)
		return n_val;

	*n_param = n_val;
	return 0;
}

static int had_setup_sample_rate(struct mwv207_audio_port *port, u32 samp_freq)
{
	int error;
	u32 n_param;

	error = had_program_n(port, samp_freq, &n_param);
	if (error) {
		pr_err("faild to programe N value %d\n", error);
		return error;
	}

	had_program_cts(port, samp_freq, port->clock, n_param);

	return error;
}

static void had_init_audio_dip(struct mwv207_audio_port *port,
			       struct snd_pcm_substream *substream)
{
	int channels;
	struct snd_pcm_runtime *runtime;
	u8 ca, threshold, conf0, conf1, layout;
	u32 value = 0;

	runtime = substream->runtime;

	threshold = 0x40;
	runtime->hw.fifo_size = threshold * 32;

	conf0 = HDMI_AHB_DMA_CONF0_SW_FIFO_RST_MASK |
	    HDMI_AHB_DMA_CONF0_INSERT_PCUV |
	    HDMI_AHB_DMA_CONF0_INCR8 | HDMI_AHB_DMA_CONF0_BURST_MODE;

	channels = substream->runtime->channels;
	if (channels <= 2) {
		ca = 0x00;
		layout = HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_LAYOUT0;
		conf1 = 0x03;
	} else {
		ca = default_hdmi_channel_config[runtime->channels - 2].ca;
		layout = HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_LAYOUT1;
		conf1 =
		    default_hdmi_channel_config[runtime->channels - 2].conf1;
	}

	had_write_hdmi_tx_reg(port, 0x3603, threshold);
	had_write_hdmi_tx_reg(port, 0x3600, conf0);
	had_write_hdmi_tx_reg(port, 0x3616, conf1);

	value = had_read_hdmi_tx_reg(port, 0x1063);
	value &= (u32) (~HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_MASK);
	value |= (layout & HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_MASK);
	had_write_hdmi_tx_reg(port, 0x1063, value);

	value = had_read_hdmi_tx_reg(port, 0x1025);
	value &= (u32) (~HDMI_FC_AUDICONF0_CC_MASK);
	value |= (((channels - 1) << HDMI_FC_AUDICONF0_CC_OFFSET) &
		  HDMI_FC_AUDICONF0_CC_MASK);
	had_write_hdmi_tx_reg(port, 0x1025, value);

	had_write_hdmi_tx_reg(port, 0x1027, ca);
}

static int had_prepare_port(struct mwv207_audio_port *port,
			    struct snd_pcm_substream *substream)
{
	int error;

	if (port->use_fake_dma)
		return 0;

#if IS_ENABLED(CONFIG_SND_PCM_ELD)
	error = snd_pcm_hw_constraint_eld(substream->runtime, port->eld);
	if (error) {
		pr_err("failed to setup pcm eld\n");
		goto out;
	}
#endif

	had_init_audio_dip(port, substream);

	error = had_setup_sample_rate(port, substream->runtime->rate);
	if (error) {
		pr_err("failed to set pcm clk regenerator\n");
		goto out;
	}

	set_auds_sample(port, substream->runtime->rate);

	had_init_dma_interrupt(port);
out:
	return error;
}

static void had_init_port_data_config(struct mwv207_audio_port *port,
				      struct snd_pcm_substream *substream)
{
	unsigned int period, rate;
	long sec;
	unsigned long nsecs;

	port->dma_addr = substream->runtime->dma_addr;
	port->channels = substream->runtime->channels;
	port->buf_size = snd_pcm_lib_buffer_bytes(substream);
	port->buf_period = snd_pcm_lib_period_bytes(substream);

	if (!port->use_fake_dma)
		return;

	period = substream->runtime->period_size;
	rate = substream->runtime->rate;
	sec = period / rate;
	period %= rate;
	nsecs = div_u64((u64) period * 1000000000UL + rate - 1, rate);

	port->period_time = ktime_set(sec, nsecs);

}

static int had_pcm_open(struct snd_pcm_substream *substream)
{
	int error;
	unsigned long flags;
	struct mwv207_audio_port *port;
	struct snd_pcm_runtime *runtime;
	port = snd_pcm_substream_chip(substream);

	runtime = substream->runtime;
	runtime->hw = had_pcm_hardware;

	error = snd_pcm_limit_hw_rates(runtime);
	if (error < 0)
		return error;

	error = snd_pcm_hw_constraint_integer(runtime,
					      SNDRV_PCM_HW_PARAM_PERIODS);
	if (error < 0)
		return error;

	error = snd_pcm_hw_constraint_minmax(runtime,
					     SNDRV_PCM_HW_PARAM_BUFFER_SIZE,
					     0, substream->dma_buffer.bytes);
	if (error < 0)
		return error;

	spin_lock_irqsave(&port->lock, flags);
	port->stream_info.substream = substream;
	port->stream_info.substream_refcount++;
	spin_unlock_irqrestore(&port->lock, flags);

	return error;
}

static void had_uninit_dma_interrupt(struct mwv207_audio_port *port)
{

	had_write_hdmi_tx_reg(port, 0x189,
			      HDMI_IH_MUTE_AHBDMAAUD_STAT0_MASK);

	had_write_hdmi_tx_reg(port, 0x3614, HDMI_AHB_DMA_MASK_MASK);

	had_write_hdmi_tx_reg(port, 0x109,
			      HDMI_IH_AHBDMAAUD_STAT0_MASK);
}

static int had_pcm_close(struct snd_pcm_substream *substream)
{
	unsigned long flags;
	struct mwv207_audio_port *port;
	port = snd_pcm_substream_chip(substream);

	spin_lock_irqsave(&port->lock, flags);
	port->stream_info.substream_refcount--;
	while (port->stream_info.substream_refcount > 0) {
		spin_unlock_irqrestore(&port->lock, flags);
		cpu_relax();
		spin_lock_irqsave(&port->lock, flags);
	}
	port->stream_info.substream = NULL;
	spin_unlock_irqrestore(&port->lock, flags);

	had_uninit_dma_interrupt(port);

	hrtimer_cancel(&port->fake_dma);

	return 0;
}

static int had_pcm_hw_params(struct snd_pcm_substream *substream,
			     struct snd_pcm_hw_params *params)
{
	return snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(params));
}

static int had_pcm_hw_free(struct snd_pcm_substream *substream)
{
	return snd_pcm_lib_free_pages(substream);
}

static int had_pcm_prepare(struct snd_pcm_substream *substream)
{
	struct mwv207_audio_port *port;
	unsigned long flags;
	int error;
	port = snd_pcm_substream_chip(substream);

	spin_lock_irqsave(&port->lock, flags);
	port->use_fake_dma = !port->enabled;
	spin_unlock_irqrestore(&port->lock, flags);

	if (port->use_fake_dma) {
		hrtimer_start(&port->fake_dma, ms_to_ktime(4),
			      HRTIMER_MODE_REL);
	} else {
		hrtimer_cancel(&port->fake_dma);
	}

	switch (substream->runtime->format) {
	case SNDRV_PCM_FORMAT_S24:
		break;
	default:
		pr_err("unsupported pcm foramt %d on audio port %d\n",
		       substream->runtime->format, port->id);
		return -EINVAL;
	}

	had_init_port_data_config(port, substream);
	error = had_prepare_port(port, substream);
	if (error)
		pr_err("failed to prepare audio port %d\n", port->id);

	return error;
}

static void had_setup_dma_addr(struct mwv207_audio_port *port, u32 start, u32 stop)
{
	had_write_hdmi_tx_reg(port, 0x3604,
			      (start & 0xff));
	had_write_hdmi_tx_reg(port, 0x3605,
			      ((start >> 8) & 0xff));
	had_write_hdmi_tx_reg(port, 0x3606,
			      ((start >> 16) & 0xff));
	had_write_hdmi_tx_reg(port, 0x3607,
			      ((start >> 24) & 0xff));

	had_write_hdmi_tx_reg(port, 0x3608, (stop & 0xff));
	had_write_hdmi_tx_reg(port, 0x3609,
			      ((stop >> 8) & 0xff));
	had_write_hdmi_tx_reg(port, 0x360A,
			      ((stop >> 16) & 0xff));
	had_write_hdmi_tx_reg(port, 0x360B,
			      ((stop >> 24) & 0xff));
}

static void had_setup_dma_buf_addr(struct mwv207_audio_port *port)
{
	u32 start, stop;
	u32 period, offset;

	offset = port->buf_offset;
	period = port->buf_period;

	start = port->dma_addr + offset;
	stop = start + period - 1;

	had_setup_dma_addr(port, start, stop);

	offset += period;
	if (offset >= port->buf_size)
		offset = 0;
	port->buf_offset = offset;
}

static void had_start_dma_xfer(struct mwv207_audio_port *port)
{

	had_write_hdmi_tx_reg(port, 0x3614,
			      (u32) (~HDMI_AHB_DMA_DONE));

	had_write_hdmi_tx_reg(port, 0x3601,
			      HDMI_AHB_DMA_START_START_MASK);
}

static void had_start_dma(struct mwv207_audio_port *port)
{
	had_setup_dma_buf_addr(port);
	had_start_dma_xfer(port);
}

static int had_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
	int error = 0;
	unsigned long flags;
	struct mwv207_audio_port *port;
	port = snd_pcm_substream_chip(substream);

	spin_lock_irqsave(&port->lock, flags);
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		port->buf_offset = 0;
		port->running = true;
		if (!port->use_fake_dma)
			had_start_dma(port);

		substream->runtime->delay = substream->runtime->period_size;
		break;

	case SNDRV_PCM_TRIGGER_SUSPEND:
	case SNDRV_PCM_TRIGGER_STOP:
		port->running = false;
		if (!port->use_fake_dma)
			had_stop_dma(port);

		break;
	default:
		error = -EINVAL;
		break;
	}
	spin_unlock_irqrestore(&port->lock, flags);

	return error;
}

static snd_pcm_uframes_t had_pcm_pointer(struct snd_pcm_substream *substream)
{
	struct mwv207_audio_port *port;

	port = substream->private_data;

	return bytes_to_frames(substream->runtime, port->buf_offset);
}

static const struct snd_pcm_ops had_pcm_ops = {
	.open = had_pcm_open,
	.close = had_pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = had_pcm_hw_params,
	.hw_free = had_pcm_hw_free,
	.prepare = had_pcm_prepare,
	.trigger = had_pcm_trigger,
	.pointer = had_pcm_pointer,
};

static int mwv207_create_jack(struct mwv207_audio_port *port, struct snd_pcm *pcm)
{
	char hdmi_str[32];
	int err;

	snprintf(hdmi_str, sizeof(hdmi_str), "HDMI/DP,pcm=%d", pcm->device);

	err = snd_jack_new(port->jcard->card, hdmi_str,
			   SND_JACK_AVOUT, &port->jack, true, false);
	if (err < 0)
		return err;

	if (port->jack == NULL)
		pr_info("sound jack is not supported.");

	return 0;
}

static inline bool hdmi_is_enabled(struct mwv207_audio_port *port)
{
	u32 state;
	u32 enabled = 1 << 0;

	state = had_read_video_if_reg(port, VIDEO_IF_CTRL);

	return !!(state & enabled);
}

static void had_jack_switch(struct mwv207_audio_port *port, bool need_jack)
{
	if (need_jack && !port->has_jack)
		snd_jack_report(port->jack, SND_JACK_AVOUT);
	if (!need_jack && port->has_jack)
		snd_jack_report(port->jack, 0);
	port->has_jack = need_jack;
}

static void had_jack_work(struct work_struct *work)
{
	struct mwv207_audio_port *port = container_of(to_delayed_work(work),
						   struct mwv207_audio_port,
						   jack_work);

	mutex_lock(&port->mutex);
	if (port->enabled) {
		BUG_ON(!port->cfg.hw_connected);
		had_jack_switch(port, true);
	} else {
		bool new_state = port->cfg.hw_connected ?
		    hdmi_is_enabled(port) : false;
		had_jack_switch(port, new_state);
	}
	mutex_unlock(&port->mutex);
}

static void had_substream_put(struct mwv207_audio_port *port)
{
	unsigned long flags;

	spin_lock_irqsave(&port->lock, flags);
	port->stream_info.substream_refcount--;
	spin_unlock_irqrestore(&port->lock, flags);
}

static struct snd_pcm_substream *had_substream_get(struct mwv207_audio_port *port)
{
	struct snd_pcm_substream *substream;
	unsigned long flags;

	spin_lock_irqsave(&port->lock, flags);
	substream = port->stream_info.substream;
	if (substream)
		port->stream_info.substream_refcount++;
	spin_unlock_irqrestore(&port->lock, flags);

	return substream;
}

static enum hrtimer_restart had_port_fake_dma_worker(struct hrtimer *timer)
{
	struct mwv207_audio_port *port =
	    container_of(timer, struct mwv207_audio_port, fake_dma);
	struct snd_pcm_substream *substream;
	unsigned long flags;
	int elapse = 0;

	spin_lock_irqsave(&port->lock, flags);
	if (port->running) {
		u32 offset;

		offset = port->buf_offset + port->buf_period;
		if (offset >= port->buf_size)
			offset = 0;
		port->buf_offset = offset;
		elapse = 1;
	}
	spin_unlock_irqrestore(&port->lock, flags);

	substream = had_substream_get(port);
	if (substream) {
		if (elapse)
			snd_pcm_period_elapsed(substream);
		had_substream_put(port);
	}

	hrtimer_forward_now(timer, port->period_time);
	return HRTIMER_RESTART;
}

static void had_port_init(struct mwv207_audio_card *jcard)
{
	int port_id;
	struct mwv207_audio_port *port;

	jcard->port_nr = 0x3;

	for_each_port(jcard, port_id) {
		port = &jcard->port[port_id];
		port->jcard = jcard;
		port->dev = jcard->dev;
		port->id = port_id;
		port->running = false;
		port->enabled = false;
		port->has_jack = false;
		port->is_dp = port_id > 3 ? true : false;
		port->stream_info.substream_refcount = 0;

		spin_lock_init(&port->lock);
		mutex_init(&port->mutex);

		hrtimer_init(&port->fake_dma, CLOCK_MONOTONIC,
			     HRTIMER_MODE_REL);
		port->fake_dma.function = had_port_fake_dma_worker;

		port->cfg.eld = NULL;
		port->cfg.clock = 0;
		port->cfg.enabled = false;
		port->cfg.hw_connected = false;
		INIT_DELAYED_WORK(&port->jack_work, had_jack_work);

		port->active = false;
	}
}

static void had_private_free(struct snd_card *card)
{
	int port_id;
	struct mwv207_audio_card *jcard;

	jcard = card->private_data;

	for_each_port(jcard, port_id) {
		struct mwv207_audio_port *port = &jcard->port[port_id];

		if (port->fake_dma.function)
			hrtimer_cancel(&port->fake_dma);

		cancel_delayed_work_sync(&port->jack_work);
	}
}

static struct pci_dev *had_get_pci_dev(struct device *dev)
{
	struct pci_dev *pdev;

	pdev = to_pci_dev(dev);
	if (!pdev) {
		pr_err("failed to get jmgpu pci device\n");
		return ERR_PTR(-ENODEV);
	}

	return pdev;
}

int mwv207_audio_init(struct mwv207_device *jdev)
{
	int error, port_id;
	struct snd_card *card;
	struct snd_pcm *pcm;
	struct pci_dev *pdev;
	struct mwv207_audio *jaudio;
	struct mwv207_audio_card *jcard;
	struct mwv207_audio_port *port;
	pdev = had_get_pci_dev(jdev->dev);
	if (IS_ERR(pdev)) {
		pr_err("failed to create mwv207 audio\n");
		return 1;
	}

	jaudio = devm_kzalloc(&pdev->dev, sizeof(struct mwv207_audio), GFP_KERNEL);
	if (!jaudio) {
		pr_err("failed to allocate mem for jmgpu audio\n");
		return 1;
	}

	error = snd_card_new(&pdev->dev, hdmi_card_index,
			     hdmi_card_id, THIS_MODULE,
			     sizeof(struct mwv207_audio_card), &card);
	if (error) {
		pr_err("failed to create snd card, error = %d\n", error);
		return 1;
	}

	strcpy(card->driver, DRIVER_NAME);
	strcpy(card->shortname, DRIVER_NAME);
	strcpy(card->longname, "Jingjia Micro HDMI/eDP Audio");
	card->private_free = had_private_free;

	jcard = card->private_data;
	jcard->card = card;
	jcard->dev = &pdev->dev;
	jcard->phys_base = pci_resource_start(pdev, 0x1);
	jcard->phys_len = pci_resource_len(pdev, 0x1);

	jcard->mmio_base = devm_ioremap(jcard->dev, jcard->phys_base,
					jcard->phys_len);
	if (!jcard->mmio_base) {
		pr_err("failed to map io at range %#llx to %#llx\n",
		       jcard->phys_base, jcard->phys_len);
		goto err_out;
	}

	had_port_init(jcard);

	for_each_port(jcard, port_id) {
		port = &jcard->port[port_id];
		error = snd_pcm_new(card, DRIVER_NAME, port_id,
				    0x1, 0x0, &pcm);
		if (error) {
			pr_err("snd_pcm_new returned error %d\n", error);
			goto err_out;
		}

		pcm->private_data = port;
		pcm->info_flags = 0;
		port->pcm = pcm;
		snprintf(pcm->name, sizeof(pcm->name), "%s %d",
			 card->shortname, port_id);

		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &had_pcm_ops);

		snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
						      &pdev->dev, 128 * 1024,
						      1024 * 1024);

		if (mwv207_create_jack(port, pcm))
			goto err_out;
	}

	jaudio->jcard = jcard;

	error = snd_card_register(card);
	if (error) {
		pr_err("failed to call snd_card_register %d\n", error);
		goto err_out;
	}

	dev_info(jcard->dev, "jmgpu audio create done\n");

	jdev->audio = jaudio;

	return 0;

err_out:
	snd_card_free(card);
	return 1;
}

void mwv207_audio_destroy(struct mwv207_audio *jaudio)
{
	if (!jaudio)
		return;

	snd_card_free(jaudio->jcard->card);
}

static void had_disable_port(struct mwv207_audio_port *port)
{
	struct snd_pcm_substream *substream;
	unsigned int delayed_time;
	unsigned long flags;

	spin_lock_irqsave(&port->lock, flags);
	if (!port->enabled) {
		spin_unlock_irqrestore(&port->lock, flags);
		return;
	}
	had_stop_dma(port);
	port->enabled = false;
	spin_unlock_irqrestore(&port->lock, flags);

	substream = had_substream_get(port);
	if (substream) {
		snd_pcm_stop_xrun(substream);
		had_substream_put(port);
	}
	delayed_time = port->cfg.hw_connected ? 3000 : 0;
	schedule_delayed_work(&port->jack_work, msecs_to_jiffies(delayed_time));
}

static void had_enable_port(struct mwv207_audio_port *port)
{
	struct snd_pcm_substream *substream;
	unsigned long flags;

	spin_lock_irqsave(&port->lock, flags);
	if (!port->active || port->enabled) {
		spin_unlock_irqrestore(&port->lock, flags);
		return;
	}
	port->enabled = true;
	spin_unlock_irqrestore(&port->lock, flags);

	substream = had_substream_get(port);
	if (substream) {
		snd_pcm_stop_xrun(substream);
		had_substream_put(port);
	}

	schedule_delayed_work(&port->jack_work, msecs_to_jiffies(50));
}

static void had_process_port_switch(struct mwv207_audio_port *port)
{
	struct port_cfg *cfg = &port->cfg;
	if (port->is_dp)
		return;

	port->active = cfg->enabled;

	if (cfg->enabled) {
		if (!cfg->eld)
			return;
		port->clock = cfg->clock;
		memcpy(port->eld, cfg->eld, sizeof(port->eld));
		had_enable_port(port);
	} else {
		port->clock = 0;
		memset(port->eld, 0, sizeof(port->eld));
		had_disable_port(port);
	}
}

void mwv207_audio_enable(struct mwv207_audio *jaudio, u32 port_idx,
			const char *eld, int clock,
			bool enabled, bool hw_connected)
{
	struct mwv207_audio_port *port;

	port = &jaudio->jcard->port[port_idx];

	mutex_lock(&port->mutex);

	port->cfg.eld = eld;
	port->cfg.clock = clock;
	port->cfg.enabled = enabled;
	port->cfg.hw_connected = hw_connected;

	had_process_port_switch(port);

	pr_info("audio enable");

	mutex_unlock(&port->mutex);
}

irqreturn_t mwv207_hdmi_audio_irq_handler(int irq, void *dev_id)
{
	u32 stat;
	struct mwv207_audio_port *port;
	struct mwv207_output *output;
	struct snd_pcm_substream *substream;

	if (!dev_id)
		return IRQ_NONE;

	output = dev_id;
	port = &output->audio->jcard->port[output->idx];

	stat = had_read_hdmi_tx_reg_raw(port, 0x109);
	if (!stat)
		return IRQ_NONE;

	had_write_hdmi_tx_reg_raw(port, 0x109, stat);

	substream = had_substream_get(port);
	if (!substream)
		return IRQ_NONE;

	if (stat & HDMI_IH_AHBDMAAUD_STAT0_DONE) {
		if (snd_pcm_playback_empty(substream))
			snd_pcm_stop_xrun(substream);

		spin_lock(&port->lock);
		if (port->enabled && port->running)
			had_start_dma(port);
		spin_unlock(&port->lock);

		snd_pcm_period_elapsed(substream);
	}

	had_substream_put(port);
	return IRQ_HANDLED;
}

int mwv207_audio_suspend(struct mwv207_device *jdev)
{
	struct mwv207_audio *jaudio = jdev->audio;
	struct mwv207_audio_card *jcard = jaudio->jcard;
	int port_id;

	snd_power_change_state(jcard->card, SNDRV_CTL_POWER_D3cold);
	for_each_port(jcard, port_id) {
		struct mwv207_audio_port *port = &jcard->port[port_id];

		hrtimer_cancel(&port->fake_dma);
		snd_pcm_suspend_all(port->pcm);
		cancel_delayed_work_sync(&port->jack_work);
		had_disable_port(port);
	}

	return 0;
}

int mwv207_audio_resume(struct mwv207_device *jdev)
{
	struct mwv207_audio *jaudio = jdev->audio;
	struct mwv207_audio_card *jcard = jaudio->jcard;
	int port_id;

	snd_power_change_state(jcard->card, SNDRV_CTL_POWER_D0);

	for_each_port(jcard, port_id) {
		struct mwv207_audio_port *port = &jcard->port[port_id];
		had_enable_port(port);
	}

	return 0;
}