/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include "mwv207_panel.h"
#include "mwv207_backlight.h"
#include <drm/drm_crtc.h>
#include <drm/drm_print.h>
#include <linux/backlight.h>
#include <linux/delay.h>
#include <linux/err.h>

#define MWV207_GPIO_BASE_ADDR (0x009aa000)
#define MWV207_GPIO_MULTIPLEX_BASE_ADDR (0x009b0910)

#define MWV207_GPIO_MULTIPLEX8_15 (MWV207_GPIO_MULTIPLEX_BASE_ADDR + 0x4)
#define MWV207_GPIO_MULTIPLEX16_23 (MWV207_GPIO_MULTIPLEX_BASE_ADDR + 0x8)
#define MWV207_GPIO_OUT_L (MWV207_GPIO_BASE_ADDR + 0x4)
#define MWV207_GPIO_DIR_L (MWV207_GPIO_BASE_ADDR + 0x8)

static DEFINE_MUTEX(panel_lock);
static LIST_HEAD(panel_list);

void mwv207_panel_init(struct mwv207_panel *panel) {
        INIT_LIST_HEAD(&panel->list);
}

int mwv207_panel_add(struct mwv207_panel *panel) {
        mutex_lock(&panel_lock);
        list_add_tail(&panel->list, &panel_list);
        mutex_unlock(&panel_lock);

        return 0;
}

void mwv207_panel_remove(struct mwv207_panel *panel) {
        mutex_lock(&panel_lock);
        list_del_init(&panel->list);
        mutex_unlock(&panel_lock);
}

static void panel_edp_wait(ktime_t start_ktime, unsigned int min_ms) {
        ktime_t now_ktime, min_ktime;

        if (!min_ms)
                return;

        min_ktime = ktime_add(start_ktime, ms_to_ktime(min_ms));
        now_ktime = ktime_get_boottime();

        if (ktime_before(now_ktime, min_ktime))
                msleep(ktime_to_ms(ktime_sub(min_ktime, now_ktime)) + 1);
}

static int edp_pwr_gpio_config(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);
        mwv207_panel_pwr *edp_pwr = jpanel->panel_pwr;

        if (!edp_pwr) {
                DRM_DEBUG_DRIVER("invalid para!");
                return -1;
        }

        mutex_lock(edp_pwr->mutex);
        if (edp_pwr->pnl_pwr_en == 0xF) {

                mwv207_panel_modify(panel, MWV207_GPIO_MULTIPLEX8_15,
                                    0b11 << 28, 0 << 28);
                mwv207_panel_modify(panel, MWV207_GPIO_DIR_L,
                                    0x1 << 0xF, 0);
        } else if (edp_pwr->pnl_pwr_en == 0x11) {

                mwv207_panel_modify(panel, MWV207_GPIO_MULTIPLEX16_23,
                                    0b11 << 4, 0 << 4);
                mwv207_panel_modify(panel, MWV207_GPIO_DIR_L,
                                    0x1 << 0x11, 0);
        }
        mutex_unlock(edp_pwr->mutex);

        DRM_DEBUG_DRIVER("mwv207L panel gpio config, pnl_pwr_en:%u",
                         edp_pwr->pnl_pwr_en);
        return 0;
}

static int edp_pwr_en_switch(struct mwv207_panel *panel, int on) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);
        mwv207_panel_pwr *edp_pwr = jpanel->panel_pwr;
        bool current_state = false;

        if (!edp_pwr) {
                DRM_DEBUG_DRIVER("invalid para!");
                return -1;
        }

        current_state = edp_pwr->ops->panel_pwr_state(panel);
        if (current_state == on) {
                DRM_DEBUG_DRIVER("mwv207: skipping dp_pwr_en(%d): %s!",
                                 edp_pwr->pnl_pwr_en, on ? "ON" : "OFF");
                return 0;
        }

        edp_pwr->ops->panel_pwr_gpio_config(panel);
        mutex_lock(edp_pwr->mutex);
        mwv207_panel_modify(panel, MWV207_GPIO_OUT_L,
                            0x1 << edp_pwr->pnl_pwr_en,
                            on ? 0x1 << edp_pwr->pnl_pwr_en : 0);
        mutex_unlock(edp_pwr->mutex);

        DRM_DEBUG_DRIVER("mwv207: panel bl_pwr_en(%d): %s!",
                         edp_pwr->pnl_pwr_en, on ? "ON" : "OFF");
        return 0;
}

static bool edp_pwr_state(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);
        mwv207_panel_pwr *edp_pwr = jpanel->panel_pwr;
        bool ret = false;

        if (!edp_pwr) {
                DRM_DEBUG_DRIVER("invalid para!");
                return -1;
        }

        mutex_lock(edp_pwr->mutex);
        if (mwv207_panel_readb(panel, MWV207_GPIO_OUT_L) &
            (0x1 << edp_pwr->pnl_pwr_en)) {
                ret = true;
        } else {
                ret = false;
        }
        mutex_unlock(edp_pwr->mutex);

        return ret;
}

static const mwv207_panel_pwr_ops edp_pwr_ops = {
    .panel_pwr_en_switch = edp_pwr_en_switch,
    .panel_pwr_gpio_config = edp_pwr_gpio_config,
    .panel_pwr_state = edp_pwr_state};

static void mwv207_panel_set_edp_hw_para(jms_panel *jpanel,
                                         mwv207_panel_pwr *panel_pwr) {
        struct mwv207_device *jdev = jpanel->jdev;

        if (!jdev->lite) {
                panel_pwr->pnl_pwr_en = (jpanel->key_idx == 6)
                                            ? 0xF
                                            : 0x11;
        } else {
                panel_pwr->pnl_pwr_en = 0xF;
        }
        panel_pwr->ops = &edp_pwr_ops;
}

static int mwv207_panel_pwr_init(struct drm_device *ddev, jms_panel *jpanel,
                                 u32 conn_type) {
        mwv207_panel_pwr *panel_pwr =
            devm_kzalloc(ddev->dev, sizeof(mwv207_panel_pwr), GFP_KERNEL);

        if (!panel_pwr) {
                DRM_DEBUG_DRIVER("panel pwr init failed");
                return -ENOMEM;
        }

        if (conn_type == DRM_MODE_CONNECTOR_eDP) {
                mwv207_panel_set_edp_hw_para(jpanel, panel_pwr);
                panel_pwr->mutex = &jpanel->jdev->gpio_lock;
                jpanel->panel_pwr = panel_pwr;
        }
        return 0;
}

static void mwv207_panel_pwr_fini(jms_panel *jpanel) {
        if (!jpanel->panel_pwr)
                return;
        jpanel->panel_pwr->pnl_pwr_en = 0;
        jpanel->panel_pwr->ops = NULL;
}

static void mwv207_panel_set_pwr(struct mwv207_panel *panel, bool pwr_state) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);
        mwv207_panel_pwr *panel_pwr = NULL;

        if (!jpanel->panel_pwr)
                return;

        panel_pwr = jpanel->panel_pwr;
        if (panel_pwr->ops && panel_pwr->ops->panel_pwr_en_switch)
                panel_pwr->ops->panel_pwr_en_switch(panel, pwr_state);
}

static int panel_prepare(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);

        DRM_DEBUG_DRIVER("panel prepare");

        if (jpanel->conn_type == DRM_MODE_CONNECTOR_eDP) {
                panel_edp_wait(jpanel->unprepared_time,
                               jpanel->edp_delay.unprepare);
                mwv207_panel_set_pwr(panel, true);
                msleep(jpanel->edp_delay.hpd_reliable);
                jpanel->prepared_time = ktime_get_boottime();
        } else {
                mwv207_panel_set_pwr(panel, true);
        }

        return 0;
}

static int panel_enable(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);

        DRM_DEBUG_DRIVER("panel enable");

        if (jpanel->conn_type == DRM_MODE_CONNECTOR_eDP) {
                msleep(jpanel->edp_delay.enable);
                panel_edp_wait(jpanel->prepared_time,
                               jpanel->edp_delay.prepare_to_enable);
        }

        if (jpanel->bl_dev)
                backlight_enable(jpanel->bl_dev);

        return 0;
}

static int panel_disable(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);

        DRM_DEBUG_DRIVER("panel disable");

        if (jpanel->bl_dev)
                backlight_disable(jpanel->bl_dev);

        if (jpanel->conn_type == DRM_MODE_CONNECTOR_eDP)
                msleep(jpanel->edp_delay.disable);

        return 0;
}

static int panel_unprepare(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);

        DRM_DEBUG_DRIVER("panel unprepare");

        mwv207_panel_set_pwr(panel, false);

        if (jpanel->conn_type == DRM_MODE_CONNECTOR_eDP)
                jpanel->unprepared_time = ktime_get_boottime();

        return 0;
}

static int panel_get_timings(struct mwv207_panel *panel,
                             unsigned int num_timings,
                             struct display_timing *timings) {

        DRM_DEBUG_DRIVER("panel get timings");
        return 0;
}

static const struct mwv207_panel_funcs panel_funcs = {
    .prepare = panel_prepare,
    .unprepare = panel_unprepare,
    .enable = panel_enable,
    .disable = panel_disable,
    .get_timings = panel_get_timings,
};

int mwv207_kms_panel_init(struct drm_device *ddev, struct mwv207_device *jdev,
                          struct mwv207_panel **panel, u32 key_idx,
                          u32 conn_type) {
        jms_panel *jpanel;

        jpanel = devm_kzalloc(ddev->dev, sizeof(jms_panel), GFP_KERNEL);
        if (!jpanel)
                return -ENOMEM;

        jpanel->jdev = jdev;
        jpanel->conn_type = conn_type;
        jpanel->ddev = ddev;
        mwv207_panel_init(&jpanel->base);
        jpanel->base.dev = ddev->dev;
        jpanel->base.funcs = &panel_funcs;
        mwv207_panel_add(&jpanel->base);
        *panel = &jpanel->base;

        if (conn_type == DRM_MODE_CONNECTOR_eDP) {

                jpanel->edp_delay.hpd_reliable = 100;
                jpanel->edp_delay.prepare_to_enable = 200;
                jpanel->edp_delay.unprepare = 500;
                jpanel->edp_delay.enable = 80;
                jpanel->edp_delay.disable = 50;
                mwv207_panel_pwr_init(ddev, jpanel, conn_type);
                bl_init(ddev, &jpanel->bl_dev, jdev, key_idx, conn_type);
        }

        return 0;
}

int mwv207_kms_panel_remove(struct mwv207_panel *panel) {
        jms_panel *jpanel = panel_to_mwv207_panel(panel);
        int err;

        err = mwv207_panel_disable(&jpanel->base);
        if (err < 0)
                DRM_ERROR("failed to disable panel: %d\n", err);

        err = mwv207_panel_unprepare(&jpanel->base);
        if (err < 0)
                DRM_ERROR("failed to unprepare panel: %d\n", err);

        if (jpanel->conn_type == DRM_MODE_CONNECTOR_eDP) {
                if (jpanel->bl_dev)
                        bl_fini(jpanel->bl_dev);
                mwv207_panel_pwr_fini(jpanel);
        }
        mwv207_panel_remove(&jpanel->base);
        return 0;
}
