/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#ifndef MWV207_PANEL_H
#define MWV207_PANEL_H

#include "../mwv207_drv.h"
#include <drm/drm_modes.h>
#include <linux/backlight.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/list.h>

struct device_node;
struct drm_device;
struct mwv207_panel;
struct display_timing;

struct mwv207_panel_funcs {
        int (*disable)(struct mwv207_panel *panel);
        int (*unprepare)(struct mwv207_panel *panel);
        int (*prepare)(struct mwv207_panel *panel);
        int (*enable)(struct mwv207_panel *panel);
        int (*get_modes)(struct mwv207_panel *panel);
        int (*get_timings)(struct mwv207_panel *panel, unsigned int num_timings,
                           struct display_timing *timings);
};

struct mwv207_panel {
        void __iomem *mmio;
        struct drm_device *drm;
        struct device *dev;

        const struct mwv207_panel_funcs *funcs;

        struct list_head list;
};

typedef struct mwv207_dp_pwr_ops {
        int (*panel_pwr_gpio_config)(struct mwv207_panel *pannel);
        int (*panel_pwr_en_switch)(struct mwv207_panel *pannel, int on);
        bool (*panel_pwr_state)(struct mwv207_panel *panel);
} mwv207_panel_pwr_ops;

typedef struct mwv207_dp_pwr {
        const mwv207_panel_pwr_ops *ops;
        struct mutex *mutex;
        u32 pnl_pwr_en;
} mwv207_panel_pwr;

typedef struct mwv207_dp_delay {
        unsigned int hpd_reliable;
        unsigned int prepare_to_enable;
        unsigned int enable;
        unsigned int disable;
        unsigned int unprepare;
} mwv207_panel_delay;

#define panel_to_mwv207_panel(panel) container_of(panel, jms_panel, base)
typedef struct tag_mwv207_panel {
        struct mwv207_device *jdev;
        struct mwv207_panel base;
        struct drm_display_mode native_mode;
        struct backlight_device *bl_dev;
        struct drm_device *ddev;
        ktime_t prepared_time;
        ktime_t unprepared_time;
        mwv207_panel_pwr *panel_pwr;
        mwv207_panel_delay edp_delay;
        u32 key_idx;
        u32 conn_type;
} jms_panel;

static inline int mwv207_panel_disable(struct mwv207_panel *panel) {
        if (panel && panel->funcs && panel->funcs->disable)
                return panel->funcs->disable(panel);

        return panel ? -ENOENT : -EINVAL;
}

static inline int mwv207_panel_unprepare(struct mwv207_panel *panel) {
        if (panel && panel->funcs && panel->funcs->unprepare)
                return panel->funcs->unprepare(panel);

        return panel ? -ENOENT : -EINVAL;
}

static inline int mwv207_panel_prepare(struct mwv207_panel *panel) {
        if (panel && panel->funcs && panel->funcs->prepare)
                return panel->funcs->prepare(panel);

        return panel ? -ENOENT : -EINVAL;
}

static inline int mwv207_panel_enable(struct mwv207_panel *panel) {
        if (panel && panel->funcs && panel->funcs->enable)
                return panel->funcs->enable(panel);

        return panel ? -ENOENT : -EINVAL;
}

static inline int mwv207_panel_get_modes(struct mwv207_panel *panel) {
        if (panel && panel->funcs && panel->funcs->get_modes)
                return panel->funcs->get_modes(panel);

        return panel ? -ENOENT : -EINVAL;
}

static inline void mwv207_panel_writeb(struct mwv207_panel *panel, u32 reg,
                                       u32 value) {
        writel_relaxed(value, panel->mmio + reg);
}

static inline u32 mwv207_panel_readb(struct mwv207_panel *panel, u32 reg) {
        return readl(panel->mmio + reg);
}

static inline void mwv207_panel_modify(struct mwv207_panel *panel, u32 reg,
                                       u32 mask, u32 value) {
        u32 rvalue;
        rvalue = mwv207_panel_readb(panel, reg);
        rvalue = (rvalue & ~mask) | (value & mask);
        mwv207_panel_writeb(panel, reg, rvalue);
}

void mwv207_panel_init(struct mwv207_panel *panel);

int mwv207_panel_add(struct mwv207_panel *panel);
void mwv207_panel_remove(struct mwv207_panel *panel);

int mwv207_kms_panel_init(struct drm_device *ddev, struct mwv207_device *jdev,
                          struct mwv207_panel **panel, u32 key_idx,
                          u32 conn_type);

#endif
