/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include "mwv207_vi.h"
#include "mwv207_vbios.h"
#include "dc/mwv207_panel.h"

bool mwv207_output_probe(struct mwv207_device *mdev, u32 key_offset)
{
	const struct mwv207_vdat *vdat;

	if (!mdev->has_cfg)
		return true;
	vdat = mwv207_vbios_vdat(mdev,
			0x0 + key_offset + 0x0);

	if (!vdat)
		return false;

	return le32_to_cpu(*(uint32_t *)vdat->dat) != 0x0;
}

static int mwv207_output_get_cfg_edid(struct mwv207_output *output, u32 key_offset)
{
	const struct mwv207_vdat *vdat;
	struct edid *edid;

	vdat = mwv207_vbios_vdat(output->jdev,
			0x0 + key_offset + 0x2);
	if (!vdat)
		return 0;
	edid = devm_kzalloc(output->jdev->dev, 256, GFP_KERNEL);
	if (!edid) {
		dev_err(output->jdev->dev, "output edid alloc failed");
		return -ENOMEM;
	}

	memcpy(edid, vdat->dat, vdat->len);
	output->forced_edid = edid;

	return 0;
}

int mwv207_output_late_register(struct drm_connector *connector)
{
	struct mwv207_output *output = connector_to_output(connector);

    	mwv207_kms_panel_init(connector->dev,output->jdev,&output->panel,
                          output->idx,output->connector.connector_type);

	output->ddc = mwv207_i2c_create(output->jdev, output->i2c_chan);
	connector->ddc = output->ddc;
	if (!output->ddc) {
		DRM_ERROR("Failed to create i2c adapter\n");
		return -ENODEV;
	}

	return 0;
}

void mwv207_output_early_unregister(struct drm_connector *connector)
{
	struct mwv207_output *output = connector_to_output(connector);

	mwv207_i2c_destroy(output->ddc);
}

int mwv207_output_get_modes(struct drm_connector *connector)
{
	struct mwv207_output *output = connector_to_output(connector);
	struct edid *edid;
	int count;

	if (output->forced_edid)
		edid = output->forced_edid;
	else
		edid = drm_get_edid(connector, output->ddc);

	output->has_audio = drm_detect_monitor_audio(edid);
	drm_connector_update_edid_property(connector, edid);
	count = drm_add_edid_modes(connector, edid);

	if (!output->forced_edid)
		kfree(edid);

	return count;
}

void mwv207_output_set_crtc(struct drm_encoder *encoder, struct drm_crtc *crtc)
{
	struct mwv207_output *output = encoder_to_output(encoder);

	output->cur_crtc = crtc;
}

void *mwv207_output_probe_and_init(struct mwv207_device *mdev,
				u32 size, int idx,
				u32 i2c_chan, u32 key_offset)
{
	struct mwv207_output *output;
	int ret;

	if (!mwv207_output_probe(mdev, key_offset))
		return NULL;

	output = devm_kzalloc(mdev->dev, size, GFP_KERNEL);
	if (!output)
		return ERR_PTR(-ENOMEM);

	output->jdev = mdev;
	ret = mwv207_output_get_cfg_edid(output, key_offset);
	if (ret)
		return ERR_PTR(ret);

	output->idx = idx;
	output->mmio = mdev->mmio + 0x9A0000;
	output->i2c_chan = i2c_chan;
	output->key_offset = key_offset;
	output->audio = mdev->audio;

	return output;
}

int mwv207_vi_init(struct mwv207_device *jdev)
{
	int ret;

	mutex_init(&jdev->gpio_lock);
	ret = mwv207_audio_init(jdev);
	if (ret)
		return ret;
	ret = mwv207_hdmi_init(jdev);
	if (ret)
		return ret;
	ret = mwv207_edp_init(jdev);
	if (ret)
		return ret;
	ret = mwv207_vga_init(jdev);
	if (ret)
		return ret;
	ret = mwv207_dvo_init(jdev);
	if (ret)
		return ret;

	return 0;
}
