/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/io-mapping.h>
#include <drm/drm_file.h>
#include "mwv207_drm.h"
#include "mwv207_bo.h"

vm_fault_t mwv207_bo_fault_reserve_notify(struct ttm_buffer_object *bo)
{
	struct ttm_operation_ctx ctx = {
		.interruptible = true,
		.no_wait_gpu = false
	};
	struct mwv207_device *jdev;
	struct mwv207_bo *jbo;
	u64 size, offset;
	u32 domain;
	int ret;

	if (!mwv207_ttm_bo_is_mwv207_bo(bo))
		return 0;
	if (bo->resource->mem_type != TTM_PL_VRAM)
		return 0;

	jbo  = to_jbo(bo);
	jdev = ddev_to_jdev(jbo->tbo.base.dev);
	size = bo->base.size;
	offset = bo->resource->start << PAGE_SHIFT;

	jbo->flags |= 0x1;

	if ((offset + size) <= jdev->visible_vram_size)
		return 0;

	if (jbo->tbo.pin_count > 0)
		return VM_FAULT_SIGBUS;

	domain = 0x2;
retry:
	mwv207_bo_placement_from_domain(jbo, domain);
	ret = ttm_bo_validate(bo, &jbo->placement, &ctx);
	if (unlikely(ret)) {
		if (ret != -ERESTARTSYS) {
			if (domain == 0x2) {

				domain = 0x1;
				goto retry;
			}
		}
		return VM_FAULT_NOPAGE;
	}

	offset = bo->resource->start << PAGE_SHIFT;
	if (bo->resource->mem_type == TTM_PL_VRAM &&
	    (offset + size) > jdev->visible_vram_size) {
		DRM_ERROR("bo validation goes crazy");
		return VM_FAULT_SIGBUS;
	}

	return 0;
}

static void mwv207_ttm_bo_destroy(struct ttm_buffer_object *bo)
{
	struct mwv207_bo *jbo = to_jbo(bo);

	WARN_ON_ONCE(jbo->map_count > 0);
	if (bo->base.import_attach)
		drm_prime_gem_destroy(&bo->base, NULL);
	drm_gem_object_release(&jbo->tbo.base);
	kfree(jbo);
}

bool mwv207_ttm_bo_is_mwv207_bo(struct ttm_buffer_object *jbo)
{
	if (jbo->destroy == &mwv207_ttm_bo_destroy)
		return true;
	return false;
}

int mwv207_bo_kmap(struct mwv207_bo *jbo, void **ptr)
{
	bool is_iomem;
	int ret;

	if (jbo->kptr) {
		if (ptr)
			*ptr = jbo->kptr;
		return 0;
	}
  ret = ttm_bo_kmap(&jbo->tbo, 0, PFN_UP(jbo->tbo.base.size), &jbo->kmap);
	if (ret)
		return ret;
	jbo->kptr = ttm_kmap_obj_virtual(&jbo->kmap, &is_iomem);
	if (ptr)
		*ptr = jbo->kptr;
	return 0;
}

void mwv207_bo_kunmap(struct mwv207_bo *jbo)
{
	if (!jbo->kptr)
		return;
	jbo->kptr = NULL;
	ttm_bo_kunmap(&jbo->kmap);
}

void mwv207_bo_placement_from_domain(struct mwv207_bo *jbo, u32 domain)
{
	struct mwv207_device *jdev = bdev_to_jdev(jbo->tbo.bdev);
	u32 c = 0;

	jbo->placement.placement = jbo->placements;
	jbo->placement.busy_placement = jbo->placements;

	if (domain & 0x2) {
		jbo->placements[c].fpfn = 0;
		jbo->placements[c].lpfn = 0;
		jbo->placements[c].mem_type = TTM_PL_VRAM;
		jbo->placements[c].flags = 0;

		if (jbo->flags & 0x4)
			jbo->placements[c].lpfn = 0x100000000ul >> PAGE_SHIFT;

		if (jbo->flags & 0x1)
			jbo->placements[c].lpfn = jdev->visible_vram_size >> PAGE_SHIFT;
		else
			jbo->placements[c].flags |= TTM_PL_FLAG_TOPDOWN;
		c++;
	}

	if (domain & 0x4) {
		jbo->placements[c].fpfn = 0;
		jbo->placements[c].lpfn = 0;
		jbo->placements[c].mem_type = TTM_PL_TT;
		jbo->placements[c++].flags = 0;
	}

	if (domain & 0x1) {
		jbo->placements[c].fpfn = 0;
		jbo->placements[c].lpfn = 0;
		jbo->placements[c].mem_type = TTM_PL_SYSTEM;
		jbo->placements[c++].flags = 0;
	}

	if (!c) {
		jbo->placements[c].fpfn = 0;
		jbo->placements[c].lpfn = 0;
		jbo->placements[c].mem_type = TTM_PL_SYSTEM;
		jbo->placements[c++].flags = 0;
	}

	BUG_ON(c > 3);
	jbo->placement.num_placement = c;
	jbo->placement.num_busy_placement = c;
}

int mwv207_bo_create(struct mwv207_device *jdev, u64 size, u64 byte_align,
		     bool kernel, u32 domain, u32 flags, struct sg_table *sg,
		     struct dma_resv *resv, struct mwv207_bo **pbo)
{
	unsigned long page_align = roundup(byte_align, PAGE_SIZE) >> PAGE_SHIFT;
	struct mwv207_bo *jbo;
	enum ttm_bo_type type;
	int ret;

	if (kernel)
		type = ttm_bo_type_kernel;
	else if (sg)
		type = ttm_bo_type_sg;
	else
		type = ttm_bo_type_device;
	*pbo = NULL;

	if ((flags & 0x4) &&
	    !(domain & 0x2))
		return -EINVAL;

	jbo = kzalloc(sizeof(struct mwv207_bo), GFP_KERNEL);
	if (!jbo)
		return -ENOMEM;
	size = roundup(size, PAGE_SIZE);
	ret = drm_gem_object_init(&jdev->base, &jbo->tbo.base, size);
	if (unlikely(ret)) {
		kfree(jbo);
		return ret;
	}
	jbo->flags  = flags;
	jbo->domain = domain;
  jbo->jdev = jdev;
	jbo->tbo.bdev = &jdev->bdev;

	mwv207_bo_placement_from_domain(jbo, domain);

	ret = ttm_bo_init_validate(&jdev->bdev, &jbo->tbo, type,
				   &jbo->placement, page_align, !kernel,
				   sg, resv, &mwv207_ttm_bo_destroy);
	if (unlikely(ret))
		return ret;
	*pbo = jbo;
	return 0;
}

int mwv207_bo_reserve(struct mwv207_bo *jbo, bool no_intr)
{
	int ret;

	ret = ttm_bo_reserve(&jbo->tbo, !no_intr, false, NULL);
	if (unlikely(ret)) {
		if (ret != -ERESTARTSYS)
			DRM_ERROR("bo reserve failed\n");
		return ret;
	}
	return 0;
}

void mwv207_bo_unreserve(struct mwv207_bo *jbo)
{
	ttm_bo_unreserve(&jbo->tbo);
}

int mwv207_bo_kmap_reserved(struct mwv207_bo *jbo, void **ptr)
{
	bool is_iomem;
	int ret;

	if (!jbo->flags & 0x1)
		return -EPERM;

	if (jbo->kptr) {
		if (ptr)
			*ptr = jbo->kptr;
		jbo->map_count++;
		return 0;
	}
	ret = ttm_bo_kmap(&jbo->tbo, 0, jbo->tbo.base.size >> PAGE_SHIFT, &jbo->kmap);
	if (ret)
		return ret;

	jbo->kptr = ttm_kmap_obj_virtual(&jbo->kmap, &is_iomem);
	if (ptr)
		*ptr = jbo->kptr;
	jbo->map_count = 1;
	return 0;
}

void mwv207_bo_kunmap_reserved(struct mwv207_bo *jbo)
{
	if (WARN_ON_ONCE(jbo->kptr == NULL))
		return;
	jbo->map_count--;
	if (jbo->map_count > 0)
		return;
	jbo->kptr = NULL;
	ttm_bo_kunmap(&jbo->kmap);
}

void mwv207_bo_unref(struct mwv207_bo *jbo)
{
	if (jbo == NULL)
		return;

	ttm_bo_put(&jbo->tbo);
}

struct mwv207_bo *mwv207_bo_ref(struct mwv207_bo *jbo)
{
	if (jbo == NULL)
		return NULL;

	ttm_bo_get(&jbo->tbo);
	return jbo;
}

int mwv207_bo_pin_reserved(struct mwv207_bo *jbo, u32 domain)
{
	struct ttm_operation_ctx ctx = {false, false};
	int ret;

	if (jbo->tbo.pin_count) {
		ttm_bo_pin(&jbo->tbo);
		return 0;
	}

	jbo->flags |= 0x4;
	mwv207_bo_placement_from_domain(jbo, domain);
	ret = ttm_bo_validate(&jbo->tbo, &jbo->placement, &ctx);
	if (ret)
		return ret;

	ttm_bo_pin(&jbo->tbo);

	return 0;
}

int mwv207_bo_unpin_reserved(struct mwv207_bo *jbo)
{
	if (WARN_ON_ONCE(!jbo->tbo.pin_count))
		return -1;

	ttm_bo_unpin(&jbo->tbo);

	return 0;
}

int mwv207_bo_pin(struct mwv207_bo *jbo, u32 domain, u64 *gpu_addr)
{
	int ret;
  struct mwv207_device *jdev = jbo->jdev;

	ret = mwv207_bo_reserve(jbo, true);
	if (ret)
		return ret;

	ret = mwv207_bo_pin_reserved(jbo, domain);
	mwv207_bo_unreserve(jbo);
  if (ret)
    return ret;

  if (gpu_addr)
		*gpu_addr = mwv207_bo_gpu_offset(jbo);
  if (domain == 0x2)
		jdev->vram_pinned_bytes += mwv207_bo_size(jbo);
	else
		jdev->gtt_pinned_bytes += mwv207_bo_size(jbo);

  return 0;
}

int mwv207_bo_unpin(struct mwv207_bo *jbo)
{
	int ret;

	ret = mwv207_bo_unpin_reserved(jbo);
	mwv207_bo_unreserve(jbo);
  if (ret)
    return ret;

  if (!jbo->tbo.pin_count) {
		if (jbo->tbo.resource->mem_type == TTM_PL_VRAM)
			jbo->jdev->vram_pinned_bytes -= mwv207_bo_size(jbo);
		else
			jbo->jdev->gtt_pinned_bytes -= mwv207_bo_size(jbo);
	}
	return 0;
}

int mwv207_bo_wait(struct mwv207_bo *jbo, bool no_wait)
{
	struct ttm_operation_ctx ctx = {
		.interruptible = true,
		.no_wait_gpu = no_wait,
	};
	int ret;

	ret = ttm_bo_reserve(&jbo->tbo, true, no_wait, NULL);
	if (unlikely(ret)) {
		if (ret != -ERESTARTSYS && ret != -EBUSY)
			DRM_ERROR("bo reserve failed for wait\n");
		return ret;
	}
	ret = ttm_bo_wait_ctx(&jbo->tbo, &ctx);
	ttm_bo_unreserve(&jbo->tbo);
	return ret;
}

int mwv207_bo_create_pin_mapped(struct mwv207_device *jdev,
				 u64 size, u64 align, u32 domain, u32 flags,
				 struct mwv207_bo **jbop, u64 *pap, void **vap)
{
	struct mwv207_bo *jbo;
	void *va;
	u64 pa;
	int ret;

	ret = mwv207_bo_create(jdev, size, align, true,
				domain, flags, NULL, NULL, &jbo);
	if (ret)
		return ret;
	ret = mwv207_bo_pin(jbo, domain, &pa);
	if (ret)
		goto unreserve_bo;
	ret = mwv207_bo_kmap(jbo, &va);
	if (ret)
		goto unpin_bo;
	mwv207_bo_unreserve(jbo);
	if (jbop)
		*jbop = jbo;
	if (pap)
		*pap = pa;
	if (vap)
		*vap = va;
	return 0;

unpin_bo:
	mwv207_bo_unpin(jbo);
unreserve_bo:
	mwv207_bo_unreserve(jbo);
	mwv207_bo_unref(jbo);
	return ret;
}

void mwv207_bo_destroy_pin_mapped(struct mwv207_bo *jbo)
{
	int ret;

	ret = mwv207_bo_reserve(jbo, true);
	if (ret) {
		pr_warn("failed to reserve bo: %d", ret);
		return;
	}
	mwv207_bo_kunmap(jbo);
	mwv207_bo_unpin(jbo);
	mwv207_bo_unreserve(jbo);
	mwv207_bo_unref(jbo);
}

