/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/dma-fence.h>
#include <drm/drm_file.h>
#include <drm/drm_ioctl.h>
#include <drm/gpu_scheduler.h>
#include "mwv207_drv.h"
#include "mwv207_drm.h"
#include "mwv207_ctx.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))

int mwv207_sched_jobs = 32;
module_param(mwv207_sched_jobs, int, 0444);
MODULE_PARM_DESC(mwv207_sched_jobs, "max number of jobs in software queue(default=32)");

struct mwv207_ctx *mwv207_ctx_lookup(struct drm_device *dev,
		struct drm_file *filp, u32 handle)
{
	struct mwv207_fpriv *fpriv = to_fpriv(filp);
	struct mwv207_ctx_mgr *mgr = &fpriv->ctx_mgr;
	struct mwv207_ctx *ctx;

	mutex_lock(&mgr->lock);
	ctx = idr_find(&mgr->handle_table, handle);
	if (ctx)
		kref_get(&ctx->refcnt);
	mutex_unlock(&mgr->lock);

	return ctx;
}

static void mwv207_ctx_entity_fini(struct mwv207_ctx *ctx)
{
	int i, j, k;

	for (i = 0; i < 0x5; i++) {
		for (j = 0; j < 0x2 + 1; j++) {
			if (ctx->centities[i][j]) {
				for (k = 0; k < mwv207_sched_jobs; k++)
					dma_fence_put(ctx->centities[i][j]->fences[k]);
				drm_sched_entity_destroy(&ctx->centities[i][j]->entity);
				kfree(ctx->centities[i][j]);
				ctx->centities[i][j] = NULL;
			}
		}
	}
}

static int mwv207_ctx_entity_init_single(struct mwv207_ctx *ctx,
		struct drm_gpu_scheduler **sched,
		struct mwv207_ctx_entity **centities, int nr)
{
  struct mwv207_ctx_entity *centity;
	int i, ret;

	for (i = 0; i < nr; i++) {
		centity = kzalloc(struct_size(centity, fences, mwv207_sched_jobs),
			 GFP_KERNEL);
		if (!centity)
      return -ENOMEM;
    mutex_init(&centity->qlock);
		ret = drm_sched_entity_init(&centity->entity,
					    DRM_SCHED_PRIORITY_NORMAL,
					    &sched[i], 1, &ctx->guilty);
		if (ret) {
			kfree(centity);
			return ret;
		}
		centities[i] = centity;
	}
	if (i) {

    centity = kzalloc(struct_size(centity, fences, mwv207_sched_jobs),
     GFP_KERNEL);
		if (!centity)
			return -ENOMEM;
    mutex_init(&centity->qlock);
		ret = drm_sched_entity_init(&centity->entity,
					    DRM_SCHED_PRIORITY_NORMAL,
					    sched, i, &ctx->guilty);
		if (ret) {
			kfree(centity);
			return ret;
		}
		centities[i] = centity;
	}

	return 0;
}

static int mwv207_ctx_entity_init(struct drm_device *dev,
				   struct mwv207_ctx *ctx)
{
	struct mwv207_device *jdev = drm_to_jdev(dev);
	int ret;
	int i;

	for (i = 0; i < 0x5; i++) {
		ret = mwv207_ctx_entity_init_single(ctx, jdev->sched[i],
						     &ctx->centities[i][0],
						     jdev->nr_pipe[i]);
		if (ret)
			goto fini;
	}

	return 0;
fini:
	mwv207_ctx_entity_fini(ctx);
	return ret;
}

static int mwv207_ctx_create(struct drm_device *dev, struct mwv207_ctx_mgr *mgr,
		u32 flags, u32 *handle)
{
	struct mwv207_ctx *ctx;
	int ret;

	ctx = kzalloc(sizeof(struct mwv207_ctx), GFP_KERNEL);
	if (ctx == NULL)
		return -ENOMEM;

	ctx->jdev = ddev_to_jdev(dev);
	kref_init(&ctx->refcnt);

	ret = mwv207_ctx_entity_init(dev, ctx);
	if (ret)
		goto out_free;

	mutex_lock(&mgr->lock);
	ret = idr_alloc(&mgr->handle_table, ctx, 1, 0, GFP_KERNEL);
	mutex_unlock(&mgr->lock);

	if (ret < 0)
		goto out_fini_entites;

	*handle = ret;
	return 0;
out_fini_entites:
	mwv207_ctx_entity_fini(ctx);
out_free:
	kfree(ctx);
	return ret;
}

static void mwv207_ctx_release(struct kref *kref)
{
	struct mwv207_ctx *ctx = container_of(kref, struct mwv207_ctx, refcnt);

	mwv207_ctx_entity_fini(ctx);
	kfree(ctx);
}

void mwv207_ctx_add_fence(struct mwv207_ctx *ctx,
			   struct drm_sched_entity *entity,
			   struct dma_fence *fence)
{
	struct mwv207_ctx_entity *centity = to_mwv207_ctx_entity(entity);
	struct dma_fence *prev;
	unsigned int idx;

	BUG_ON(!mutex_is_locked(&centity->qlock));

	idx = centity->sequence & (mwv207_sched_jobs - 1);
	prev = centity->fences[idx];

	if (likely(prev)) {
		BUG_ON(!dma_fence_is_signaled(prev));
		dma_fence_put(prev);
	}

	centity->fences[idx] = dma_fence_get(fence);
	centity->sequence++;
}

int mwv207_ctx_wait_prev_fence(struct mwv207_ctx *ctx,
				struct drm_sched_entity *entity)
{
	struct mwv207_ctx_entity *centity = to_mwv207_ctx_entity(entity);
	struct dma_fence *prev;
	unsigned int idx;

	BUG_ON(!mutex_is_locked(&centity->qlock));

	idx = centity->sequence & (mwv207_sched_jobs - 1);
	prev = centity->fences[idx];
	if (likely(prev))
		return dma_fence_wait(prev, true);

	return 0;
}

int mwv207_ctx_put(struct mwv207_ctx *ctx)
{
	return	kref_put(&ctx->refcnt, mwv207_ctx_release);
}

static int mwv207_ctx_destroy(struct drm_device *dev, struct mwv207_ctx_mgr *mgr, u32 handle)
{
	struct mwv207_ctx *ctx;

	mutex_lock(&mgr->lock);
	ctx = idr_remove(&mgr->handle_table, handle);
	mutex_unlock(&mgr->lock);

	if (ctx == NULL)
		return -EINVAL;

	mwv207_ctx_put(ctx);
	return 0;
}

int  mwv207_ctx_ioctl(struct drm_device *dev, void *data,
			    struct drm_file *filp)
{
	struct drm_mwv207_ctx *args = (struct drm_mwv207_ctx *)data;
	struct mwv207_fpriv *fpriv = to_fpriv(filp);

	if (args->resv || args->flags)
		return -EINVAL;

	switch (args->op) {
	case 0x0:
		return mwv207_ctx_create(dev, &fpriv->ctx_mgr, args->flags, &args->handle);
	case 0x1:
		return mwv207_ctx_destroy(dev, &fpriv->ctx_mgr, args->handle);
	default:
		return -EINVAL;
	}
}

int mwv207_ctx_mgr_init(struct drm_device *dev, struct mwv207_ctx_mgr *mgr)
{
	mutex_init(&mgr->lock);
	idr_init(&mgr->handle_table);
	return 0;
}

void mwv207_ctx_mgr_fini(struct drm_device *dev, struct mwv207_ctx_mgr *mgr)
{
	struct mwv207_ctx *ctx;
	uint32_t id;

	idr_for_each_entry(&mgr->handle_table, ctx, id)
		mwv207_ctx_put(ctx);

	idr_destroy(&mgr->handle_table);
	mutex_destroy(&mgr->lock);
}

static struct drm_sched_entity *mwv207_create_kentity(
		struct mwv207_device *jdev, int type)
{
	struct drm_gpu_scheduler **sched = jdev->sched[type];
	int nr_pipe = jdev->nr_pipe[type];
	struct drm_sched_entity *entity;
	int ret;

	entity = devm_kzalloc(jdev->dev,
			      sizeof(struct drm_sched_entity),
			      GFP_KERNEL);
	if (!entity)
		return NULL;

	ret = drm_sched_entity_init(entity,
				    DRM_SCHED_PRIORITY_HIGH,
				    sched, nr_pipe, NULL);
	if (ret)
		return NULL;

	return entity;
}

int mwv207_kctx_init(struct mwv207_device *jdev)
{
	if (mwv207_sched_jobs < 4) {
		pr_warn("jobs(%d) at least 4", mwv207_sched_jobs);
		mwv207_sched_jobs = 4;
	} else if (!is_power_of_2(mwv207_sched_jobs)) {
		pr_warn("jobs(%d) must be power of 2", mwv207_sched_jobs);
		mwv207_sched_jobs = roundup_pow_of_two(mwv207_sched_jobs);
	}

	jdev->dma_entity = mwv207_create_kentity(jdev, 0x4);
	jdev->blt_entity = mwv207_create_kentity(jdev, 0x3);

	if (!jdev->dma_entity || !jdev->blt_entity) {
		mwv207_kctx_fini(jdev);
		return -ENOMEM;
	}
	mutex_init(&jdev->qlock);

	return 0;
}

void mwv207_kctx_fini(struct mwv207_device *jdev)
{
	if (jdev->dma_entity)
		drm_sched_entity_destroy(jdev->dma_entity);
	if (jdev->blt_entity)
		drm_sched_entity_destroy(jdev->blt_entity);
}
