/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/dma-buf.h>
#include <drm/drm_gem_ttm_helper.h>
#include <drm/drm_utils.h>
#include <drm/drm_prime.h>
#include "mwv207_drm.h"
#include "mwv207_gem.h"
#include "mwv207_bo.h"

const struct drm_gem_object_funcs mwv207_gem_object_funcs;

static vm_fault_t mwv207_gem_fault(struct vm_fault *vmf)
{
	struct ttm_buffer_object *bo = vmf->vma->vm_private_data;
	vm_fault_t ret;

	ret = ttm_bo_vm_reserve(bo, vmf);
	if (ret)
		goto unlock_resv;

	ret = mwv207_bo_fault_reserve_notify(bo);
	if (ret)
		goto unlock_resv;

	ret = ttm_bo_vm_fault_reserved(vmf, vmf->vma->vm_page_prot, TTM_BO_VM_NUM_PREFAULT);
	if (ret == VM_FAULT_RETRY && !(vmf->flags & FAULT_FLAG_RETRY_NOWAIT))
		return ret;

unlock_resv:
	dma_resv_unlock(bo->base.resv);
	return ret;
}

static const struct vm_operations_struct mwv207_ttm_vm_ops = {
	.fault = mwv207_gem_fault,
	.open = ttm_bo_vm_open,
	.close = ttm_bo_vm_close,
	.access = ttm_bo_vm_access
};

int mwv207_gem_prime_pin(struct drm_gem_object *gem)
{
	struct mwv207_bo *jbo = mwv207_bo_from_gem(gem);
	int ret;

	ret = mwv207_bo_reserve(jbo, true);
	if (ret)
		return ret;
	ret = mwv207_bo_pin_reserved(jbo, 0x2);
	mwv207_bo_unreserve(jbo);
	return ret;
}

void mwv207_gem_prime_unpin(struct drm_gem_object *gem)
{
	struct mwv207_bo *jbo = mwv207_bo_from_gem(gem);
	int ret;

	ret = mwv207_bo_reserve(jbo, true);
	if (ret)
		return;
	ret = mwv207_bo_unpin_reserved(jbo);
	mwv207_bo_unreserve(jbo);
}

int mwv207_bo_vmap(struct mwv207_bo *jbo, struct iosys_map *map)
{
	int ret;

	if (jbo->kptr) {
		jbo->map_count++;
		goto out;
	}

	ret = ttm_bo_vmap(&jbo->tbo, &jbo->map);
	if (ret) {
		return ret;
	}
	jbo->map_count = 1;

	if (jbo->map.is_iomem)
		jbo->kptr = (void *)jbo->map.vaddr_iomem;
	else
		jbo->kptr = jbo->map.vaddr;

out:
	*map = jbo->map;
	return 0;
}

int mwv207_gem_prime_vmap(struct drm_gem_object *gem, struct iosys_map *map)
{
	struct mwv207_bo *jbo = mwv207_bo_from_gem(gem);
	int ret;

	ret = mwv207_bo_reserve(jbo, false);
	if (ret)
		return ret;

	jbo->flags |= 0x1;
	ret = mwv207_bo_pin_reserved(jbo, 0x2);
	if (unlikely(ret)) {
		drm_err(gem->dev, "pin %p for vmap failed\n", jbo);
		goto unreserve_bo;
	}

	ret = mwv207_bo_vmap(jbo, map);
	if (ret) {
		drm_err(gem->dev, "ttm bo vmap failed\n");
		goto unpin_bo;
	}

	mwv207_bo_unreserve(jbo);
	return ret;

unpin_bo:
	ttm_bo_unpin(&jbo->tbo);
unreserve_bo:
	mwv207_bo_unreserve(jbo);
	return ret;
}

void mwv207_bo_vunmap(struct mwv207_bo *jbo)
{
	if (WARN_ON_ONCE(jbo->kptr == NULL))
		return;
	jbo->map_count--;
	if (jbo->map_count > 0)
		return;
	jbo->kptr = NULL;
	ttm_bo_vunmap(&jbo->tbo, &jbo->map);
}

void mwv207_gem_prime_vunmap(struct drm_gem_object *gem, struct iosys_map *map)
{
	struct mwv207_bo *jbo = mwv207_bo_from_gem(gem);
	int ret;

	ret = mwv207_bo_reserve(jbo, false);
	if (ret)
		return;

	mwv207_bo_vunmap(jbo);
	ttm_bo_unpin(&jbo->tbo);
	mwv207_bo_unreserve(jbo);
}

static int mwv207_gem_object_open(struct drm_gem_object *obj,
				  struct drm_file *file_priv)
{
	return 0;
}

static void mwv207_gem_object_close(struct drm_gem_object *obj,
				    struct drm_file *file_priv)
{
}

static struct sg_table *
mwv207_gem_prime_get_sg_table(struct drm_gem_object *obj)
{
	struct mwv207_bo *jbo = mwv207_bo_from_gem(obj);

	return drm_prime_pages_to_sg(obj->dev, jbo->tbo.ttm->pages,
				     jbo->tbo.ttm->num_pages);
}

struct drm_gem_object *
mwv207_gem_prime_import_sg_table(struct drm_device *dev,
				 struct dma_buf_attachment *attach,
				 struct sg_table *sg)
{
	struct dma_resv *resv = attach->dmabuf->resv;
	struct mwv207_device *jdev = ddev_to_jdev(dev);
	struct mwv207_bo *jbo;
	int ret;

	dma_resv_lock(resv, NULL);
	ret = mwv207_bo_create(jdev, attach->dmabuf->size, PAGE_SIZE, false,
				0x4, 0, sg, resv, &jbo);
	dma_resv_unlock(resv);
	if (ret)
		return ERR_PTR(ret);

	jbo->tbo.base.funcs = &mwv207_gem_object_funcs;

	return &jbo->tbo.base;
}

static void mwv207_gem_object_free(struct drm_gem_object *gobj)
{
	struct mwv207_bo *jbo = mwv207_bo_from_gem(gobj);

	mwv207_bo_unref(jbo);
}

struct dma_buf *mwv207_gem_prime_export(struct drm_gem_object *gobj,
					int flags)
{
	return drm_gem_prime_export(gobj, flags);
}

const struct drm_gem_object_funcs mwv207_gem_object_funcs = {
	.free           = mwv207_gem_object_free,
	.open           = mwv207_gem_object_open,
	.close          = mwv207_gem_object_close,
	.export         = mwv207_gem_prime_export,
	.pin            = mwv207_gem_prime_pin,
	.unpin          = mwv207_gem_prime_unpin,
	.get_sg_table   = mwv207_gem_prime_get_sg_table,
	.vmap           = mwv207_gem_prime_vmap,
	.vunmap         = mwv207_gem_prime_vunmap,
	.mmap           = drm_gem_ttm_mmap,
	.vm_ops         = &mwv207_ttm_vm_ops,
};

static int mwv207_gem_create(struct mwv207_device *jdev,
		u64 size, u64 align, u32 preferred_domain,
		u32 flags, struct drm_gem_object **gobj)
{
	struct mwv207_bo *jbo;
	int ret;

retry:

	ret = mwv207_bo_create(jdev, size, align, false, preferred_domain,
			       flags, NULL, NULL, &jbo);
	if (ret) {
		if (ret != -ERESTARTSYS) {
			if (flags & 0x1) {
				flags &= ~0x1;
				goto retry;
			}
			if (preferred_domain == 0x2) {
				preferred_domain |= 0x1;
				goto retry;
			}
			DRM_DEBUG("Failed to allocate GEM object (%lld, %d, %llu, %d)\n",
				  size, preferred_domain, align, ret);
		}
		return ret;
	}

	jbo->tbo.base.funcs = &mwv207_gem_object_funcs;

	*gobj = mwv207_gem_from_bo(jbo);
	return 0;
}

int mwv207_gem_dumb_create(struct drm_file *file, struct drm_device *dev,
		     struct drm_mode_create_dumb *args)
{
	struct mwv207_device *jdev = ddev_to_jdev(dev);
	struct drm_gem_object *gobj;
	u32 handle;
	int ret;

	args->pitch = ALIGN(args->width * DIV_ROUND_UP(args->bpp, 8), 64);
	args->size = args->pitch * args->height;
	ret = mwv207_gem_create(jdev, args->size, 0x10000,
			0x2,
			0x1,
			&gobj);
	if (ret)
		return ret;

	ret = drm_gem_handle_create(file, gobj, &handle);
	drm_gem_object_put(gobj);
	if (ret)
		return ret;

	args->handle = handle;
	return 0;
}

void mwv207_gem_free_object(struct drm_gem_object *gobj)
{
	struct mwv207_bo *jbo;

	if (!gobj)
		return;

	jbo = mwv207_bo_from_gem(gobj);
	mwv207_bo_unref(jbo);
}

int  mwv207_gem_create_ioctl(struct drm_device *dev, void *data,
			    struct drm_file *filp)
{
	struct mwv207_device *jdev = ddev_to_jdev(dev);
	union drm_mwv207_gem_create *args = data;
	struct drm_gem_object *gobj;
	int ret;

	if (args->in.size == 0)
		return -EINVAL;
	if (args->in.alignment & (args->in.alignment - 1))
		return -EINVAL;
	if (args->in.preferred_domain & ~0x7)
		return -EINVAL;
	if (args->in.flags & ~0x7)
		return -EINVAL;

	ret = mwv207_gem_create(jdev, args->in.size, args->in.alignment,
			args->in.preferred_domain,
			args->in.flags, &gobj);
	if (ret)
		return ret;

	ret = drm_gem_handle_create(filp, gobj, &args->out.handle);
	drm_gem_object_put(gobj);

	return ret;
}

int  mwv207_gem_mmap_ioctl(struct drm_device *dev, void *data,
			    struct drm_file *filp)
{
	union drm_mwv207_gem_mmap *args = data;
	struct drm_gem_object *obj;

	if (args->in.pad)
		return -EINVAL;

	obj = drm_gem_object_lookup(filp, args->in.handle);
	if (!obj)
		return -ENOENT;
	args->out.offset = mwv207_bo_mmap_offset(obj);
	drm_gem_object_put(obj);

	return 0;
}

int  mwv207_gem_wait_ioctl(struct drm_device *dev, void *data,
			    struct drm_file *filp)
{
	struct drm_mwv207_gem_wait *args = (struct drm_mwv207_gem_wait *)data;
	long timeout;
	bool write;
	int ret;

	if (args->op & ~(0x2 | 0x1))
		return -EINVAL;

	write = args->op & 0x2;
	timeout = drm_timeout_abs_to_jiffies(args->timeout);

	ret = drm_gem_dma_resv_wait(filp, args->handle, write, timeout);
	if (ret == -ETIME)
		ret = timeout ? -ETIMEDOUT : -EBUSY;

	return ret;
}
