/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/kernel.h>
#include <linux/pci.h>

#include "mwv207_drv.h"
#include "mwv207_irq.h"
#include "mwv207_vbios.h"
#include "mwv207_db.h"

static inline u32 hwinfo_read(struct mwv207_device *jdev, u32 reg)
{
	u32 regbase = jdev->hw.is_pf ? 0x25FC00 : 0xF000;
	return jdev_read(jdev, regbase + reg);
}

static u64 mwv207_vram_size(struct mwv207_device *jdev)
{
	int nr = 0;
	u64 size;

	switch (jdev_read(jdev, 0x9B0190) & 0x7) {
	case 0:
		size = 2048;
		break;
	case 1:
		size = 1024;
		break;
	case 2:
		size = 4096;
		break;
	case 3:
		size = 8192;
		break;
	default:
		pr_warn("mwv207: ddr size default to 256MB");
		return 256UL * 1024 * 1024;
	}

	if (jdev->lite) {
		nr += jdev_read(jdev, 0xd101bc) & 0x1;
		nr += jdev_read(jdev, 0xd11cbc) & 0x1;
	} else {
		nr += (jdev_read(jdev, 0x9702f4) & 0x02000000) ? 1 : 0;
		nr += (jdev_read(jdev, 0x9802f4) & 0x02000000) ? 1 : 0;
	}
	if (!nr) {
		nr = 1;
		pr_warn("mwv207: ddr cnt default to 1");
	}

	return size * 1024 * 1024 * nr;
}

static void mwv207_hwinfo_parse(struct mwv207_device *jdev)
{
	struct mwv207_hwinfo *hw = &jdev->hw;
	u32 val;

  val = jdev_read(jdev,0xD10000);
  hw->vram_type = (val & 0x20) ? "LPDDR4" : "DDR4";

	hw->vram_size = mwv207_vram_size(jdev);
	BUG_ON(!hw->vram_size);

	val = hwinfo_read(jdev, 0x4);
	hw->nr_3d_cluster = val & 0x1f;
	hw->nr_2d_cluster = (val >> 5) & 0x1f;
	hw->nr_fus_cluster = (val >> 10) & 0x1f;

  hw->nr_stream_processor = 256;
  hw->vram_bit_wide = 64;

	hw->nr_pipe[0x0] = 0x2;
	hw->nr_pipe[0x3] = 0x1;
	hw->nr_pipe[0x4] = 0x1;
	hw->nr_pipe[0x1] = 0x2;
	hw->nr_pipe[0x2] = 0x1;

	hw->vf_sel = hwinfo_read(jdev, 0x8) & 0xf;
	hw->vram_offset = hwinfo_read(jdev, 0xC);
	hw->vram_offset <<= 20;

  WARN(!hw->nr_pipe[0x0], "no 3d engine!");
	WARN(!hw->nr_pipe[0x3], "no 2d engine!");
	WARN(!hw->nr_pipe[0x4], "no dma engine!");

	dev_info(jdev->dev, "running on %s", hw->is_pf ? "pf" : "vf");
	dev_info(jdev->dev, "3d cluster:  %d", hw->nr_3d_cluster);
	dev_info(jdev->dev, "2d cluster:  %d", hw->nr_2d_cluster);
	dev_info(jdev->dev, "fus cluster: %d", hw->nr_fus_cluster);
  dev_info(jdev->dev, "dma engine:  %d", hw->nr_pipe[0x4]);
	dev_info(jdev->dev, "decoder:     %d", hw->nr_pipe[0x1]);
	dev_info(jdev->dev, "encoder:     %d", hw->nr_pipe[0x2]);
}

int mwv207_hw_init(struct mwv207_device *jdev)
{
	struct pci_dev *pdev = to_pci_dev(jdev->dev);
	resource_size_t size;

	if (!jdev->mmio) {
		dev_err(jdev->dev, "failed to map mmio");
		return -ENOMEM;
	}

	size = pci_resource_len(pdev, 0x1);
	jdev->hw.is_pf = size != SZ_2M;

	if (jdev->is_haps && !jdev->hw.is_pf && !jdev->isr_poll) {
		pr_warn("warning, haps100 no irq, enable isr_poll");
		jdev->isr_poll = true;
	}

	if (jdev->is_haps && jdev->hw.is_pf && !jdev->isr_poll)
		pr_warn("warning, running on haps's pf with irq enabled");

	jdev->gtt_size = SZ_2G;
	jdev->ap_base = pci_resource_start(pdev, 0x0);
	jdev->ap_size = pci_resource_len(pdev, 0x0);
	if (!jdev->ap_base || !jdev->ap_size) {
		dev_err(jdev->dev, "error, pci <%04x:%04x> bar%d no resource!",
			pdev->vendor, pdev->device, 0x0);
		return -ENODEV;
	}

	mwv207_hwinfo_parse(jdev);

	jdev->visible_vram_size = min(jdev->hw.vram_size, jdev->ap_size);

	dev_info(jdev->dev, "%lluM VRAM, %lluM aperture, %lluM visible VRAM",
		 jdev->hw.vram_size >> 20,
		 jdev->ap_size >> 20,
		 jdev->visible_vram_size >> 20);

	jdev->ap_vaddr = devm_ioremap_wc(jdev->dev, jdev->ap_base,
					 jdev->visible_vram_size);
	if (!jdev->ap_vaddr) {
		dev_err(jdev->dev, "failed to mmap aperture vram");
		return -ENOMEM;
	}

	mwv207_db_add(jdev, DRM_MWV207_3D_CLUSTER_NR, jdev->hw.nr_3d_cluster);
	mwv207_db_add(jdev, DRM_MWV207_2D_CLUSTER_NR, jdev->hw.nr_2d_cluster);
	mwv207_db_add(jdev, DRM_MWV207_FUS_CLUSTER_NR,
		      jdev->hw.nr_fus_cluster);
	mwv207_db_add(jdev, DRM_MWV207_VRAM_SIZE, jdev->hw.vram_size);
	mwv207_db_add(jdev, DRM_MWV207_GTT_SIZE, jdev->gtt_size);
	mwv207_db_add(jdev, DRM_MWV207_VISIBLE_VRAM_SIZE,
		      jdev->visible_vram_size);
	mwv207_db_add(jdev, DRM_MWV207_FAMILY, 0x1100);

  return 0;
}

void mwv207_hw_fini(struct mwv207_device *jdev)
{
	mwv207_irq_fini(jdev);
	if (jdev->hw.is_pf)
		mwv207_vbios_fini(jdev);
}
