/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/kernel.h>
#include "mwv207_monitor.h"
#include "mwv207_drv.h"
#include "mwv207_bo.h"

static int monitor_hb_period = 4000;
module_param(monitor_hb_period, int, 0444);
MODULE_PARM_DESC(monitor_hb_period, "heartbeat period, default 4000ms");

struct mwv207_monitor {
  struct mwv207_bo *bo;
  u64 gpu_addr;
  u32 *vaddr;
  struct delayed_work hb_work;
  int hb_period;
  u32 heartbeat;
  bool inited;
};

static inline void monitor_update(struct mwv207_monitor *m, int idx, u32 val) {
  WRITE_ONCE(m->vaddr[idx], val);
}

static inline void monitor_get(struct mwv207_monitor *m, int idx, u32 *val) {
  *val = READ_ONCE(m->vaddr[idx]);
}

void mwv207_monitor_update_pipe_usage(struct mwv207_device *jdev, u8 pipe_type,
                                      u8 pipe_id, u32 usage) {
  struct mwv207_monitor *m = jdev->monitor;
  int idx;

  if (unlikely((!m || !m->inited))) return;

  BUG_ON(pipe_type >= 0x5);
  BUG_ON(pipe_id >= 0x2);
  BUG_ON(pipe_id >= jdev->nr_pipe[pipe_type]);

  switch (pipe_type) {
    case 0x0:
      idx = 0x19;
      break;
    case 0x3:
      idx = 0x18;
      break;
    case 0x2:
      idx = 0x1A + pipe_id;
      break;
    case 0x1:
      idx = 0x1B + pipe_id;
      break;
    default:
      return;
  }

  monitor_update(m, idx, usage);
}

void mwv207_monitor_update_vram_used(struct mwv207_device *jdev) {
  struct mwv207_monitor *m = jdev->monitor;

  if (unlikely((!m || !m->inited))) return;

  monitor_update(m, 0x8,
                 (u32)(atomic64_read(&jdev->vram_used_bytes) >> 10));
  monitor_update(m, 0x9,
                 (u32)(atomic64_read(&jdev->visible_vram_used_bytes) >> 10));
}

static void mwv207_monitor_info_init(struct mwv207_monitor *m) {
  u32 version = 1 << 24 | DRIVER_MAJOR << 16 | DRIVER_MINOR << 8;

  monitor_update(m, 0x0, 0);
  monitor_update(m, 0x1, version);
  monitor_update(m, 0x18, 0xffffffffu);
  monitor_update(m, 0x19, 0xffffffffu);
  monitor_update(m, 0x1A, 0xffffffffu);
  monitor_update(m, 0x1B, 0xffffffffu);
  monitor_update(m, 0x1C, 0xffffffffu);
  monitor_update(m, 0x1D, 0xffffffffu);
}

static void monitor_hb_work(struct work_struct *work) {
  struct mwv207_monitor *m =
      container_of(to_delayed_work(work), struct mwv207_monitor, hb_work);

  monitor_update(m, 0x0, ++m->heartbeat);
  schedule_delayed_work(&m->hb_work, msecs_to_jiffies(m->hb_period));
}

int mwv207_monitor_init(struct mwv207_device *jdev) {
  struct mwv207_monitor *m;
  int ret;

  m = devm_kzalloc(jdev->dev, sizeof(*m), GFP_KERNEL);
  if (!m) return -ENOMEM;
  jdev->monitor = m;

  ret = mwv207_bo_create_pin_mapped(jdev, SZ_4K, SZ_4K, 0x2,
                                    0x1, &m->bo,
                                    &m->gpu_addr, (void **)&m->vaddr);
  if (ret) {
    pr_err("failed to allocate monitor bo");
    return ret;
  }
  m->inited = true;

  BUG_ON(m->gpu_addr != 0);

  mwv207_monitor_info_init(m);

  if (monitor_hb_period < 100) monitor_hb_period = 100;
  if (monitor_hb_period > 7000) monitor_hb_period = 7000;

  m->hb_period = monitor_hb_period;

  INIT_DELAYED_WORK(&m->hb_work, monitor_hb_work);

  schedule_delayed_work(&m->hb_work, msecs_to_jiffies(m->hb_period));

  return 0;
}

void mwv207_monitor_fini(struct mwv207_device *jdev) {
  cancel_delayed_work_sync(&jdev->monitor->hb_work);
  jdev->monitor->inited = false;
  mwv207_bo_destroy_pin_mapped(jdev->monitor->bo);
}
