/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/moduleparam.h>
#include "mwv207_drv.h"
#include "mwv207_drm.h"
#include "mwv207_sched.h"
#include "mwv207_bo.h"
#include "mwv207_ctx.h"
#include "mwv207_db.h"
#include "mwv207_devfreq.h"
#include "mwv207_monitor.h"

static int dump_hang_job = 1;
module_param(dump_hang_job, int, 0644);
MODULE_PARM_DESC(dump_hang_job, "dump time out job");

static unsigned long pipe_usage_period = 50;
module_param(pipe_usage_period, ulong, 0444);
MODULE_PARM_DESC(pipe_usage_period, "pipe usage calculation period in ms");

struct mwv207_sched_info {
	struct mwv207_pipe *(*create)(struct mwv207_device *jdev,
			     const struct mwv207_pipe_info *info);
	unsigned int hw_submission;
	unsigned int hang_limit;
	long timeout;
	struct mwv207_pipe_info pipes[0x2];
};

static struct mwv207_sched_info mwv207_sched_pipe_info[0x5] = {
	[0x0] = {
		.create = mwv207_pipe_3d_create,
		.hw_submission = 30,
		.hang_limit = 5,
		.timeout = 2000,
		.pipes = {
			{ "mwv207_3d0", 0, 0x900000, 64, 0 },
			{ "mwv207_3d1", 1, 0x910000, 64, 0 },
		}
	},
	[0x1] = {
		.create = mwv207_pipe_dec_create,
		.hw_submission = 1,
		.hang_limit = 5,
		.timeout = 2000,
		.pipes = {
			{ "mwv207_dec0", 0, 0x930000, 104, 5 },
			{ "mwv207_dec1", 1, 0x940000, 108, 6 },
		}
	},
	[0x2] = {
		.create = mwv207_pipe_enc_create,
		.hw_submission = 1,
		.hang_limit = 5,
		.timeout = 2000,
		.pipes = {
			{ "mwv207_enc0", 0, 0x920000, 96, 3 },
		}
	},
	[0x3] = {
		.create = mwv207_pipe_2d_create,
		.hw_submission = 30,
		.hang_limit = 5,
		.timeout = 2000,
		.pipes = {
			{ "mwv207_2d0", 0, 0x968000, 72, 1 },
		}
	},
	[0x4] = {
		.create = mwv207_pipe_dma_create,
		.hw_submission = 1,
		.hang_limit = 5,
		.timeout = 2000,
		.pipes = {
			{ "mwv207_dma0", 0, 0x9D0000, 88, 32 },
		}
	},
};

static u32 mwv207_pipe_update_usage(struct mwv207_pipe *pipe)
{
	struct mwv207_pipe_usage *usage = &pipe->usage;
	unsigned long total_time;
	unsigned long busy_time;
	unsigned long delta;
	u32 record;
	ktime_t last;
	ktime_t now;

	now = ktime_get();
	last = usage->last_update_time;

	delta = ktime_to_ms(ktime_sub(now, usage->period_start));

	if (usage->busy_count > 0)
		usage->busy_time += ktime_sub(now, last);
	else
		usage->idle_time += ktime_sub(now, last);

	usage->last_update_time = now;

	if (delta > pipe_usage_period) {
		usage->last_busy_time = usage->busy_time;
		usage->last_idle_time = usage->idle_time;
		usage->busy_time = 0;
		usage->idle_time = 0;
		usage->period_start = now;
	}

	total_time = ktime_to_ns(ktime_add(usage->last_busy_time,
					   usage->last_idle_time));
	busy_time = ktime_to_ns(usage->last_busy_time);

	if (total_time == 0)
		record = 100;
	else
		record = (int)div64_u64(busy_time, total_time / 100);

	mwv207_monitor_update_pipe_usage(pipe->jdev, pipe->type,
					  pipe->id, record);

	return record;
}

u32 mwv207_pipe_get_usage(struct mwv207_device *jdev,
			   u8 pipe_type, u8 pipe_id)
{
	struct mwv207_pipe *pipe;
	unsigned long flags;
	u32 usage;

	BUG_ON(pipe_type >= 0x5);
	BUG_ON(pipe_id >= 0x2);

	if (pipe_id >= jdev->nr_pipe[pipe_type])
		return 0xffffffffu;
	if (!jdev->sched[pipe_type][pipe_id])
		return 0xffffffffu;

	pipe = to_mwv207_sched(jdev->sched[pipe_type][pipe_id])->pipe;

	spin_lock_irqsave(&pipe->usage.lock, flags);

	usage = mwv207_pipe_update_usage(pipe);

	spin_unlock_irqrestore(&pipe->usage.lock, flags);

	return usage;
}

void mwv207_pipe_record_busy(struct mwv207_pipe *pipe)
{
	unsigned long flags;

	spin_lock_irqsave(&pipe->usage.lock, flags);

	mwv207_pipe_update_usage(pipe);

	pipe->usage.busy_count++;

	spin_unlock_irqrestore(&pipe->usage.lock, flags);
}

void mwv207_pipe_record_idle(struct mwv207_pipe *pipe)
{
	unsigned long flags;

	spin_lock_irqsave(&pipe->usage.lock, flags);

	mwv207_pipe_update_usage(pipe);

	WARN_ON(--pipe->usage.busy_count < 0);

	spin_unlock_irqrestore(&pipe->usage.lock, flags);
}

static enum drm_mwv207_info_key sched_type_to_key(int type)
{
	switch (type) {
	case 0x0:
		return DRM_MWV207_ACTIVE_3D_NR;
	case 0x1:
		return DRM_MWV207_ACTIVE_DEC_NR;
	case 0x2:
		return DRM_MWV207_ACTIVE_ENC_NR;
	case 0x3:
		return DRM_MWV207_ACTIVE_2D_NR;
	case 0x4:
		return DRM_MWV207_ACTIVE_DMA_NR;
	default:
		BUG_ON(1);
		return 0;
	}
}

struct mwv207_job *mwv207_job_alloc(void)
{
	struct mwv207_job *mjob;

	mjob = kzalloc(sizeof(*mjob), GFP_KERNEL);
	if (!mjob)
		return NULL;

	kref_init(&mjob->refcount);
	INIT_LIST_HEAD(&mjob->tvblist);
	return mjob;
}

static void mwv207_job_fini(struct kref *kref)
{
	struct mwv207_job *mjob = container_of(kref, struct mwv207_job, refcount);
	struct mwv207_tvb *mtvb;
	struct mwv207_submit_cmd_dat *dat;
	int i;

	dma_fence_put(mjob->fence_in);
	if (mjob->base.s_fence)
		drm_sched_job_cleanup(&mjob->base);

	mwv207_for_each_mtvb(mtvb, mjob) {
		for (i = 0; i < mtvb->nr_shared; ++i)
			dma_fence_put(mtvb->shared[i]);
		dma_fence_put(mtvb->excl);
		kfree(mtvb->shared);
		if (mtvb->base.bo)
			mwv207_bo_unref(to_jbo(mtvb->base.bo));
	}

	if (mjob->nr_cmd_dat) {
		for (i = 0; i < mjob->nr_cmd_dat; ++i) {
			dat = &mjob->cmd_dats[i];
			kvfree(dat->dat);
			kvfree(dat->relocs);
			mwv207_bo_unref(dat->jbo);
		}
	}

	if (mjob->ctx)
		mwv207_ctx_put(mjob->ctx);

	kvfree(mjob->cmd_dats);
	kvfree(mjob->mtvb);
	kvfree(mjob->relocs);
	kvfree(mjob->cmds);

	kfree(mjob);
}

struct mwv207_job *mwv207_job_get(struct mwv207_job *mjob)
{
	kref_get(&mjob->refcount);
	return mjob;
}

void mwv207_job_put(struct mwv207_job *mjob)
{
	kref_put(&mjob->refcount, mwv207_job_fini);
}

static void mwv207_sched_dump_job(struct mwv207_job *mjob)
{
	int lines, i;
	u32 *cmd;

	cmd = (u32 *)mjob->cmds;
	lines = (mjob->cmd_size / 4) & ~0x3;
	for (i = 0; i < lines; i += 4) {
		pr_info("0x%08x : %08x %08x %08x %08x", i,
				cmd[i], cmd[i+1], cmd[i+2], cmd[i+3]);
	}

	switch (mjob->cmd_size / 4 - lines) {
	case 0:
		return;
	case 1:
		pr_info("0x%08x : %08x", i, cmd[i]);
		break;
	case 2:
		pr_info("0x%08x : %08x %08x", i, cmd[i], cmd[i+1]);
		break;
	case 3:
		pr_info("0x%08x : %08x %08x %08x", i,
				cmd[i], cmd[i+1], cmd[i+2]);
		break;
	default:
		pr_err("should never happen");
		break;

	}
}

static struct dma_fence *mwv207_sched_dependency(struct drm_sched_job *job,
					       struct drm_sched_entity *entity)
{
	struct mwv207_job *mjob = to_mwv207_job(job);
	struct mwv207_tvb *mtvb;
	struct dma_fence *fence;
	int i;

	if (unlikely(mjob->fence_in)) {
		fence = mjob->fence_in;
		mjob->fence_in = NULL;
		if (!dma_fence_is_signaled(fence))
			return fence;
		dma_fence_put(fence);
	}

	mwv207_for_each_mtvb(mtvb, mjob) {
		if (mtvb->excl) {
			fence = mtvb->excl;
			mtvb->excl = NULL;
			if (!dma_fence_is_signaled(fence))
				return fence;
			dma_fence_put(fence);
		}

		for (i = 0; i < mtvb->nr_shared; i++) {
			if (!mtvb->shared[i])
				continue;
			fence = mtvb->shared[i];
			mtvb->shared[i] = NULL;
			if (!dma_fence_is_signaled(fence))
				return fence;
			dma_fence_put(fence);
		}
		kfree(mtvb->shared);
		mtvb->nr_shared = 0;
		mtvb->shared = NULL;
	}

	return NULL;
}

static struct dma_fence *mwv207_sched_run_job(struct drm_sched_job *job)
{
	struct mwv207_sched *sched = to_mwv207_sched(job->sched);
	struct dma_fence *fence;

	if (unlikely(job->s_fence->finished.error))
		return ERR_PTR(-ECANCELED);

	mwv207_devfreq_record_busy(sched->pipe);

	fence = sched->pipe->submit(sched->pipe, to_mwv207_job(job));
	if (IS_ERR_OR_NULL(fence))
		DRM_ERROR("%s: submit failed, errcode=%ld",
				sched->pipe->fname, PTR_ERR(fence));
	return fence;
}

static enum drm_gpu_sched_stat
mwv207_sched_timedout_job(struct drm_sched_job *job)
{
	struct mwv207_sched *sched = to_mwv207_sched(job->sched);
	struct mwv207_job *mjob = to_mwv207_job(job);
	struct drm_sched_job *timedout_job;

	pr_info("mwv207: job from '%s' to '%s', timed out!",
			mjob->comm, sched->pipe->fname);

	drm_sched_stop(&sched->base, job);

	if (dump_hang_job) {
		sched->pipe->dump_state(sched->pipe);
		mwv207_sched_dump_job(mjob);
	}

	drm_sched_increase_karma(job);

	sched->pipe->reset(sched->pipe);

	list_for_each_entry(timedout_job, &sched->base.pending_list, list)
		mwv207_devfreq_record_idle(sched->pipe);

	drm_sched_resubmit_jobs(&sched->base);

	drm_sched_start(&sched->base, true);

	return DRM_GPU_SCHED_STAT_NOMINAL;
}

static void mwv207_sched_free_job(struct drm_sched_job *job)
{
	struct mwv207_job *mjob = to_mwv207_job(job);

	mwv207_job_put(mjob);
}

static const struct drm_sched_backend_ops mwv207_sched_ops = {
	.prepare_job = mwv207_sched_dependency,
	.run_job = mwv207_sched_run_job,
	.timedout_job = mwv207_sched_timedout_job,
	.free_job = mwv207_sched_free_job,
};

static struct drm_gpu_scheduler *
mwv207_sched_create(struct mwv207_device *jdev, struct mwv207_pipe *pipe,
		     unsigned int hw_submission, unsigned int hang_limit,
		     long timeout)
{
	struct mwv207_sched *sched;
	int ret;

	if (pipe == NULL)
		return NULL;

	sched = devm_kzalloc(jdev->dev, sizeof(struct mwv207_sched), GFP_KERNEL);
	if (!sched)
		goto err;

	sched->pipe = pipe;

	ret = drm_sched_init(&sched->base, &mwv207_sched_ops, hw_submission, hang_limit,
			msecs_to_jiffies(timeout), NULL, NULL, pipe->fname, jdev->dev);
	if (ret)
		goto err;

	return &sched->base;
err:
	sched->pipe->destroy(sched->pipe);
	return NULL;
}

static void mwv207_sched_destroy(struct drm_gpu_scheduler *sched)
{
	struct mwv207_sched *msched = to_mwv207_sched(sched);

	if (sched == NULL)
		return;

	drm_sched_fini(sched);
	msched->pipe->destroy(msched->pipe);
}

static int mwv207_sched_pipe_create(struct mwv207_device *jdev, int type)
{
	struct mwv207_sched_info *info = &mwv207_sched_pipe_info[type];
	struct mwv207_pipe *pipe;
	int i;

	BUG_ON(jdev->hw.nr_pipe[type] > 0x2);

	for (i = 0; i < jdev->hw.nr_pipe[type]; i++) {
		if (!info->pipes[i].reg_base)
			break;
		pipe = info->create(jdev, &info->pipes[i]);
		if (!pipe)
			break;
		jdev->sched[type][i] = mwv207_sched_create(jdev, pipe,
							    info->hw_submission,
							    info->hang_limit,
							    info->timeout);
		if (!jdev->sched[type][i]) {
			pr_err("error, create %s failed", info->pipes[i].name);
			break;
		}
	}

	return i;
}

int mwv207_sched_suspend(struct mwv207_device *jdev)
{
	struct mwv207_sched *sched;
	int i, j;

	for (i = 0; i < 0x5; i++) {
		for (j = 0; j < 0x2; j++) {
			if (!jdev->sched[i][j])
				continue;

			sched = to_mwv207_sched(jdev->sched[i][j]);

			if (atomic_read(&jdev->sched[i][j]->hw_rq_count))
				return -EBUSY;
		}
	}

	return 0;
}

int mwv207_sched_resume(struct mwv207_device *jdev)
{
	struct mwv207_sched *sched;
	int i, j;

	for (i = 0; i < 0x5; i++) {
		for (j = 0; j < 0x2; j++) {
			if (!jdev->sched[i][j])
				continue;
			sched = to_mwv207_sched(jdev->sched[i][j]);
			sched->pipe->reset(sched->pipe);
		}
	}

	return 0;
}
int mwv207_sched_init(struct mwv207_device *jdev)
{
	int i;

	for (i = 0; i < 0x5; i++) {
		jdev->nr_pipe[i] = mwv207_sched_pipe_create(jdev, i);
		mwv207_db_add(jdev, sched_type_to_key(i), jdev->nr_pipe[i]);
	}

	return 0;
}

void mwv207_sched_fini(struct mwv207_device *jdev)
{
	int i, j;

	for (i = 0; i < 0x5; ++i) {
		for (j = 0; j < 0x2; j++) {
			if (!jdev->sched[i][j])
				continue;
			mwv207_sched_destroy(jdev->sched[i][j]);
		}
	}

}
