/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/delay.h>
#include <drm/drm.h>
#include <drm/drm_print.h>
#include <drm/drm_file.h>
#include <drm/drm_debugfs.h>
#include <drm/ttm/ttm_placement.h>
#include <drm/ttm/ttm_range_manager.h>

#include "mwv207_drv.h"
#include "mwv207_bo.h"
#include "mwv207_drm.h"
#include "mwv207_sched.h"
#include "mwv207_monitor.h"

static int mwv207_mem_visible(struct mwv207_device *jdev, struct ttm_resource *mem)
{
	if (mem->mem_type == TTM_PL_SYSTEM ||
	    mem->mem_type == TTM_PL_TT)
		return true;
	if (mem->mem_type != TTM_PL_VRAM)
		return false;

	return ((mem->start << PAGE_SHIFT) + mem->size < jdev->visible_vram_size);
}

static void mwv207_evict_flags(struct ttm_buffer_object *bo,
		struct ttm_placement *placement)
{
	struct mwv207_device *jdev;
	struct mwv207_bo *jbo;
	static const struct ttm_place placements = {
		.fpfn = 0,
		.lpfn = 0,
		.mem_type = TTM_PL_SYSTEM,
		.flags = 0
	};

	if (!mwv207_ttm_bo_is_mwv207_bo(bo)) {
		placement->placement = &placements;
		placement->busy_placement = &placements;
		placement->num_placement = 1;
		placement->num_busy_placement = 1;
		return;
	}

	jbo = to_jbo(bo);
	jdev = ddev_to_jdev(jbo->tbo.base.dev);

	if (bo->resource->mem_type == TTM_PL_VRAM) {
		if (jdev->visible_vram_size != jdev->vram_size
			&& mwv207_mem_visible(jdev, bo->resource)) {
			jbo->flags &= ~(0x1 | 0x4);
			mwv207_bo_placement_from_domain(jbo,
					0x2 | 0x4);
			BUG_ON(!(jbo->placements[0].mem_type == TTM_PL_VRAM));
			BUG_ON(!(jbo->placements[1].mem_type == TTM_PL_TT));
			jbo->placements[0].fpfn = jdev->visible_vram_size >> PAGE_SHIFT;
			jbo->placements[0].lpfn = 0;
			jbo->placement.placement = &jbo->placements[0];
			jbo->placement.busy_placement = &jbo->placements[1];
			jbo->placement.num_placement = 1;
			jbo->placement.num_busy_placement = 1;
		} else
			mwv207_bo_placement_from_domain(jbo, 0x4);
	} else
		mwv207_bo_placement_from_domain(jbo, 0x1);

	*placement = jbo->placement;
}

static int mwv207_ttm_io_mem_reserve(struct ttm_device *bdev,
				     struct ttm_resource *mem)
{
	struct mwv207_device *jdev = bdev_to_jdev(bdev);
	size_t bus_size = (size_t)mem->size;

	switch (mem->mem_type) {
	case TTM_PL_SYSTEM:
	case TTM_PL_TT:
		break;
	case TTM_PL_VRAM:
		mem->bus.offset = mem->start << PAGE_SHIFT;

		if ((mem->bus.offset + bus_size) > jdev->visible_vram_size)
			return -EINVAL;
		mem->bus.offset += jdev->vram_bar_base;
		mem->bus.is_iomem = true;
		mem->bus.caching = ttm_write_combined;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static struct ttm_tt *mwv207_ttm_tt_create(struct ttm_buffer_object *bo,
		u32 page_flags)
{
	struct mwv207_ttm_tt *gtt;

	gtt = kzalloc(sizeof(*gtt), GFP_KERNEL);
	if (!gtt)
		return NULL;

	if (ttm_sg_tt_init(&gtt->ttm, bo, page_flags, ttm_cached)) {
		kfree(gtt);
		return NULL;
	}

	return &gtt->ttm;
}

static int mwv207_ttm_tt_populate(struct ttm_device *bdev, struct ttm_tt *ttm,
				  struct ttm_operation_ctx *ctx)
{
	return ttm_pool_alloc(&bdev->pool, ttm, ctx);
}

static void mwv207_ttm_tt_unpopulate(struct ttm_device *bdev, struct ttm_tt *ttm)
{
	ttm_pool_free(&bdev->pool, ttm);
}

static int mwv207_ttm_job_submit(struct mwv207_job *mjob, struct dma_fence **fence)
{
	*fence = &mjob->base.s_fence->finished;

	mwv207_job_get(mjob);
	drm_sched_entity_push_job(&mjob->base);

	return 0;
}

static struct mwv207_job *mwv207_ttm_job_alloc(struct ttm_buffer_object *bo)
{
	struct mwv207_job *mjob = mwv207_job_alloc();
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_tvb *mtvb;
	int ret;

	if (!mjob)
		return ERR_PTR(-ENOMEM);

	dma_cmd = kmalloc(sizeof(*dma_cmd), GFP_KERNEL);
	if (!dma_cmd) {
		ret = -ENOMEM;
		goto err;
	}
	dma_cmd->src.offset = 0;
	dma_cmd->src.stride = bo->resource->size;
	dma_cmd->dst.offset = 0;
	dma_cmd->dst.stride = bo->resource->size;
	dma_cmd->width = bo->resource->size;
	dma_cmd->height = 1;

	mjob->cmds = (char *)dma_cmd;
	mjob->cmd_size = sizeof(*dma_cmd);
	mjob->engine_entity = ddev_to_jdev(bo->base.dev)->dma_entity;
	mjob->is_dma = true;

	mjob->mtvb = kzalloc(sizeof(struct mwv207_tvb), GFP_KERNEL);
	if (!mjob->mtvb) {
		ret = -ENOMEM;
		goto err;
	}
	mtvb = &mjob->mtvb[0];
	list_add_tail(&mtvb->base.head, &mjob->tvblist);

	ret = drm_sched_job_init(&mjob->base, mjob->engine_entity, NULL);
	if (ret)
		goto err;

	drm_sched_job_arm(&mjob->base);

	ret = drm_sched_job_add_implicit_dependencies(&mjob->base, &bo->base, true);
	if (ret)
		goto err;

	return mjob;
err:
	mwv207_job_put(mjob);
	return ERR_PTR(ret);
}

static int mwv207_move_vram_gtt(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_resource *new_mem)
{
	struct ttm_tt *ttm = bo->ttm;
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_job *mjob;
	struct dma_fence *fence;
	int ret;

	mjob = mwv207_ttm_job_alloc(bo);
	if (IS_ERR(mjob))
		return PTR_ERR(mjob);

	dma_cmd = (struct mwv207_dma_cmd *)mjob->cmds;
	dma_cmd->src.base = bo->resource->start << PAGE_SHIFT;
	dma_cmd->src.pg_nr_type = ((bo->resource->size >> PAGE_SHIFT) | (0ULL << 63));
	dma_cmd->dst.base = (u64)gtt->ttm.dma_address;
	dma_cmd->dst.pg_nr_type = ((ttm->num_pages) | (1ULL << 63));

	ret = mwv207_ttm_job_submit(mjob, &fence);
	if (!ret)
		ret = ttm_bo_move_accel_cleanup(bo, fence, evict, true, new_mem);

	mwv207_job_put(mjob);
	return ret;
}

static int mwv207_move_gtt_vram(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_resource *new_mem)
{
	struct ttm_tt *ttm = bo->ttm;
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_job *mjob;
	struct dma_fence *fence;
	int ret;

	mjob = mwv207_ttm_job_alloc(bo);
	if (IS_ERR(mjob))
		return PTR_ERR(mjob);

	dma_cmd = (struct mwv207_dma_cmd *)mjob->cmds;
	dma_cmd->src.base = (u64)gtt->ttm.dma_address;
	dma_cmd->src.pg_nr_type = ((ttm->num_pages) | (1ULL << 63));
	dma_cmd->dst.base = new_mem->start << PAGE_SHIFT;
	dma_cmd->dst.pg_nr_type = ((bo->resource->size >> PAGE_SHIFT) | (0ULL << 63));

	ret = mwv207_ttm_job_submit(mjob, &fence);
	if (!ret)
		ret = ttm_bo_move_accel_cleanup(bo, fence, evict, true, new_mem);

	mwv207_job_put(mjob);
	return ret;
}

static int mwv207_move_vram_vram(struct ttm_buffer_object *bo, bool evict,
				 struct ttm_operation_ctx *ctx,
				 struct ttm_resource *new_mem)
{
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_job *mjob;
	struct dma_fence *fence;
	int ret;

	mjob = mwv207_ttm_job_alloc(bo);
	if (IS_ERR(mjob))
		return PTR_ERR(mjob);

	dma_cmd = (struct mwv207_dma_cmd *)mjob->cmds;
	dma_cmd->src.base = bo->resource->start << PAGE_SHIFT;
	dma_cmd->src.pg_nr_type = ((bo->resource->size >> PAGE_SHIFT) | (0ULL << 63));
	dma_cmd->dst.base = new_mem->start << PAGE_SHIFT;
	dma_cmd->dst.pg_nr_type = ((new_mem->size >> PAGE_SHIFT) | (0ULL << 63));

	ret = mwv207_ttm_job_submit(mjob, &fence);
	if (!ret)
		ret = ttm_bo_move_accel_cleanup(bo, fence, evict, true, new_mem);

	mwv207_job_put(mjob);
	return ret;
}

static int mwv207_bo_move_blit(struct ttm_buffer_object *bo, bool evict,
			       struct ttm_operation_ctx *ctx,
			       struct ttm_resource *new_mem)
{
	if (bo->resource->mem_type == TTM_PL_TT &&
	    new_mem->mem_type == TTM_PL_VRAM)
		return mwv207_move_gtt_vram(bo, evict, ctx, new_mem);
	if (bo->resource->mem_type == TTM_PL_VRAM &&
	    new_mem->mem_type == TTM_PL_TT)
		return mwv207_move_vram_gtt(bo, evict, ctx, new_mem);
	if (bo->resource->mem_type == TTM_PL_VRAM &&
	    new_mem->mem_type == TTM_PL_VRAM)
		return mwv207_move_vram_vram(bo, evict, ctx, new_mem);
	return -EFAULT;
}

static void mwv207_update_vram_usage(struct ttm_buffer_object *bo, bool add)
{
	struct mwv207_device *jdev = bdev_to_jdev(bo->bdev);

	if (add)
		atomic64_add(bo->base.size, &jdev->vram_used_bytes);
	else
		atomic64_sub(bo->base.size, &jdev->vram_used_bytes);

	mwv207_monitor_update_vram_used(jdev);
}

static void mwv207_update_memory_usage(struct ttm_buffer_object *bo,
					int mem_type, bool add)
{
	struct mwv207_device *jdev;

	if (!mwv207_ttm_bo_is_mwv207_bo(bo))
		return;

	jdev = bdev_to_jdev(bo->bdev);

	switch (mem_type) {
	case TTM_PL_TT:
		if (add)
			atomic64_add(bo->base.size, &jdev->gtt_used_bytes);
		else
			atomic64_sub(bo->base.size, &jdev->gtt_used_bytes);
		break;
	case TTM_PL_VRAM:
		mwv207_update_vram_usage(bo, add);
		break;
	}
}

static void mwv207_bo_move_notify(struct ttm_buffer_object *bo,
				   int old_type,
				   struct ttm_resource *new_mem)
{
	mwv207_update_memory_usage(bo, old_type, false);
	if (new_mem)
		mwv207_update_memory_usage(bo, new_mem->mem_type, true);

}

static int mwv207_bo_move(struct ttm_buffer_object *bo, bool evict,
			  struct ttm_operation_ctx *ctx,
			  struct ttm_resource *new_mem,
			  struct ttm_place *hop)
{
	struct ttm_resource *old_mem = bo->resource;
	struct mwv207_bo *jbo = to_jbo(bo);
	struct mwv207_device *jdev;
	int ret, old_type;

	if (WARN_ON_ONCE(bo->pin_count > 0))
		return -EINVAL;

	if (old_mem == NULL || (old_mem->mem_type == TTM_PL_SYSTEM && bo->ttm == NULL)) {

		ttm_bo_move_null(bo, new_mem);
		return 0;
	}

	old_type = old_mem->mem_type;

	jdev = bdev_to_jdev(bo->bdev);

	if (old_mem->mem_type == TTM_PL_SYSTEM && new_mem->mem_type == TTM_PL_TT) {
		ttm_bo_move_null(bo, new_mem);
		return 0;
	}

	if (old_mem->mem_type == TTM_PL_TT &&
	    new_mem->mem_type == TTM_PL_SYSTEM) {

		return -EINVAL;
	}

	if ((old_mem->mem_type == TTM_PL_SYSTEM && new_mem->mem_type == TTM_PL_VRAM) ||
	    (old_mem->mem_type == TTM_PL_VRAM && new_mem->mem_type == TTM_PL_SYSTEM)) {
		hop->fpfn = 0;
		hop->lpfn = 0;
		hop->mem_type = TTM_PL_TT;
		hop->flags = 0;
		return -EMULTIHOP;
	}

	ret = mwv207_bo_move_blit(bo, evict, ctx, new_mem);
	if (ret == 0)
		goto moved;

	if (!mwv207_mem_visible(jdev, new_mem)
			|| !mwv207_mem_visible(jdev, old_mem)) {
		pr_err("mwv207: can't move invisible mem by cpu");
		return ret;
	}
	ret = ttm_bo_move_memcpy(bo, ctx, new_mem);
	if (ret)
		return ret;
moved:
  atomic64_add(bo->base.size, &jdev->moved_bytes);
  mwv207_bo_move_notify(bo, old_type, new_mem);
	if (!mwv207_mem_visible(jdev, new_mem))
		jbo->flags &= ~0x1;

	return 0;
}

static void mwv207_ttm_delete_mem_notify(struct ttm_buffer_object *bo)
{
}

static void mwv207_ttm_tt_destroy(struct ttm_device *bdev,
				  struct ttm_tt *ttm)
{
	struct mwv207_ttm_tt *gtt = container_of(ttm, struct mwv207_ttm_tt, ttm);

	ttm_tt_fini(ttm);
	kfree(gtt);
}

static struct ttm_device_funcs mwv207_bo_driver = {
	.ttm_tt_create = mwv207_ttm_tt_create,
	.ttm_tt_destroy = mwv207_ttm_tt_destroy,
	.ttm_tt_populate = mwv207_ttm_tt_populate,
	.ttm_tt_unpopulate = mwv207_ttm_tt_unpopulate,
	.eviction_valuable = ttm_bo_eviction_valuable,
	.evict_flags = mwv207_evict_flags,
	.move = mwv207_bo_move,
	.io_mem_reserve = mwv207_ttm_io_mem_reserve,
	.delete_mem_notify = mwv207_ttm_delete_mem_notify,
};

int mwv207_ttm_init(struct mwv207_device *jdev)
{
	int ret;

	ret = ttm_device_init(&jdev->bdev, &mwv207_bo_driver, jdev->dev,
			       jdev->base.anon_inode->i_mapping,
			       jdev->base.vma_offset_manager,
			       false, true);
	if (ret) {
		DRM_ERROR("failed to initialize buffer object driver(%d).\n", ret);
		return ret;
	}

	ret = ttm_range_man_init(&jdev->bdev, TTM_PL_TT,
				  true, SZ_2G >> PAGE_SHIFT);
	if (ret) {
		DRM_ERROR("failed to initialize GTT heap.\n");
		goto out_no_vram;
	}

	ret = ttm_range_man_init(&jdev->bdev, TTM_PL_VRAM,
				 false, jdev->vram_size >> PAGE_SHIFT);
	if (ret) {
		DRM_ERROR("failed to initialize VRAM heap.\n");
		goto out_no_gtt;
	}
	DRM_INFO("mwv207: %lldM of VRAM memory size\n", jdev->vram_size / (1024 * 1024));
	return 0;

out_no_gtt:
	ttm_range_man_fini(&jdev->bdev, TTM_PL_TT);
out_no_vram:
	ttm_device_fini(&jdev->bdev);
	return ret;
}

void mwv207_ttm_fini(struct mwv207_device *jdev)
{
	ttm_range_man_fini(&jdev->bdev, TTM_PL_VRAM);
	ttm_range_man_fini(&jdev->bdev, TTM_PL_TT);
	ttm_device_fini(&jdev->bdev);
	DRM_INFO("mwv207: ttm finalized\n");
}
