/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/time.h>
#include <linux/delay.h>
#include "mwv207_drv.h"
#include "mwv207_vbios.h"

#define MWV207_PLL_BIT_MASK(_x)   ((u32)(((u32)0x01U) << (_x)))
#define MWV207_PLL_BITS_MASK(_start, _end) ((u32)(((0xFFFFFFFFUL << (_start))) & (0xFFFFFFFFUL >> (31U - (u32)(_end)))))

#define MWV207REG_PLL_STATE_BASE_ADDR(_id)            (0x9B0000 + 0xA00 + 0x100 * ((_id)/6) + 0x20 * ((_id) % 6))
#define MWV207REG_PLL_FBDIV_REFDIV_STATE(_id)         (MWV207REG_PLL_STATE_BASE_ADDR(_id) + 0x00)
#define MWV207REG_PLL_FRAC_STATE(_id)                 (MWV207REG_PLL_STATE_BASE_ADDR(_id) + 0x04)
#define MWV207REG_PLL_POSTDIV3_0_STATE(_id)           (MWV207REG_PLL_STATE_BASE_ADDR(_id) + 0x08)
#define MWV207REG_PLL_POSTDIV4_STATE(_id)             (MWV207REG_PLL_STATE_BASE_ADDR(_id) + 0x0C)
#define MWV207REG_PLL_SSMOD_STATE(_id)                (MWV207REG_PLL_STATE_BASE_ADDR(_id) + 0x10)

#define PLL_RECFG_POSTDIV_OFFSET(_port)            (8*((_port)%4))
#define PLL_RECFG_POSTDIV_MASK(_port)              (MWV207_PLL_BITS_MASK(0, 6) << (8*((_port)%4)))

static int mwv207_vcmd_get_pll(struct mwv207_device *jdev, int pll_idx, u32 *kfreq)
{
	u32 port = 0;
	u32 reg_val_0 = 0;
	u32 reg_val_1 = 0;
	u32 reg_val_2 = 0;
	u64 temp = 0;
	u32 refdiv;
	u32 fbintdiv;
	u32 fbracdiv;
	u32 postdiv1;
	u32 postdiv2;

	if (pll_idx == 0)
		port = 1;

	mutex_lock(&jdev->vbios.vcmd_lock);
	reg_val_0 = jdev_read(jdev, MWV207REG_PLL_FBDIV_REFDIV_STATE(pll_idx));
	reg_val_1 = jdev_read(jdev, MWV207REG_PLL_FRAC_STATE(pll_idx));
	reg_val_2 = jdev_read(jdev, MWV207REG_PLL_POSTDIV3_0_STATE(pll_idx));
	mutex_unlock(&jdev->vbios.vcmd_lock);

	refdiv = (reg_val_0 & MWV207_PLL_BITS_MASK(0,5)) >> 0x0;
	fbintdiv = (reg_val_0 & MWV207_PLL_BITS_MASK(16,27)) >> 0x10;
	fbracdiv = (reg_val_1 & MWV207_PLL_BITS_MASK(0,23)) >> 0x0;
	postdiv1 = 0x4;
	postdiv2 =
		(reg_val_2 & PLL_RECFG_POSTDIV_MASK(port)) >>
		PLL_RECFG_POSTDIV_OFFSET(port);

	if (!refdiv || !postdiv1 || !postdiv2) {
		dev_err(jdev->dev, "idx %u: illegal divisor\n", pll_idx);
		return -EINVAL;
	}

	temp = (100000u) * 100000ull / refdiv;
	temp = temp * fbintdiv + ((temp * fbracdiv) >> 24);
	temp /= postdiv1;
	temp /= postdiv2;
	temp = DIV_ROUND_CLOSEST_ULL(temp, 100000);
	*kfreq = (u32) temp;

	return 0;
}

static inline int mwv207_vcmd_idle(struct mwv207_device *jdev)
{
	u32 state;

	state = jdev_read(jdev, 0x9B0144) >> 28;

	return state == 0x0;
}

static inline int mwv207_vcmd_wait_idle(struct mwv207_device *jdev)
{
	int i;

	for (i = 0; i < 1000 && !mwv207_vcmd_idle(jdev); i++)
		usleep_range(1000, 1001);
	return mwv207_vcmd_idle(jdev) ? 0 : -ETIMEDOUT;
}

static inline void mwv207_vcmd_req(struct mwv207_device *jdev, int cmd,
		int idx, u32 value, int enable)
{
	u32 req;

	req  = cmd << 28;
	req |= idx << 24;
	req |= value << 1;
	req |= enable ? 1 : 0;

	jdev_write(jdev, 0x9B0140, req);
}

static inline void mwv207_vcmd_start(struct mwv207_device *jdev)
{
	jdev_write(jdev, 0x9B0144, 0x1 << 28);
}

static inline int mwv207_vcmd_result(struct mwv207_device *jdev)
{
	return (jdev_read(jdev, 0x9B0144) >> 24) & 0xf;
}

static int mwv207_vcmd_execute(struct mwv207_device *jdev, int cmd,
		int idx, u32 value, int enable)
{
	int ret;

	mutex_lock(&jdev->vbios.vcmd_lock);

	ret = mwv207_vcmd_wait_idle(jdev);
	if (ret)
		goto unlock;

	mwv207_vcmd_req(jdev, cmd, idx, value, enable);

	mwv207_vcmd_start(jdev);

	ret = mwv207_vcmd_wait_idle(jdev);
	if (ret)
		goto unlock;

	ret = mwv207_vcmd_result(jdev);
	switch (ret) {
	case 0:
		goto unlock;
	case 1:
		pr_err("unsupport pll type!)");
		break;
	case 2:

		pr_info_once("pll set, invalid freq!");
		break;
	case 3:
		pr_err("pll busy!");
		break;
	case 4:
		pr_err("pll set, invalid argument!");
		break;
	case 9:
		pr_err("PLL REQ cmd code unsupport!");
		break;
	default:
		pr_err("unknown pll error! (code 0x%x)", ret);
		break;
	}
	ret = -EIO;
unlock:
	mutex_unlock(&jdev->vbios.vcmd_lock);
	return ret;
}

int mwv207_vbios_set_pll(struct mwv207_device *jdev, int pll_idx, u32 kfreq)
{
	if (pll_idx >= MWV207_PLL_COUNT)
		return -EINVAL;

	return mwv207_vcmd_execute(jdev, 0x1, pll_idx,
			kfreq, kfreq ? 1 : 0);
}

int mwv207_vbios_get_pll(struct mwv207_device *jdev,
				   int pll_idx, u32 *kfreq)
{
	if (pll_idx >= MWV207_PLL_COUNT)
		return -EINVAL;

	return mwv207_vcmd_get_pll(jdev, pll_idx, kfreq);
}

int mwv207_vbios_get_vram_pll(struct mwv207_device *jdev, u32 *freq, u32 *rate){
  u32 val, i;
  u32 ddr4_rate[8][2] = {
    {400,1600}, {465,1866}, {532,2133}, {600,2400},
    {655,2666}, {732,2933}, {800,3200}, {931,3733}
  };

  mwv207_vbios_get_pll(jdev, MWV207_PLL_DDR, &val);
  val /= 1000;
  for (i=0; i<8; i++){
    if ((val >= ddr4_rate[i][0] - 5) && (val <= ddr4_rate[i][0] + 5)){
      *freq = ddr4_rate[i][0];
      *rate = ddr4_rate[i][1];
    }
  }

  return 0;
}
